import pytest

from nortl.core.constructs import Condition, Fork
from nortl.core.engine import CoreEngine
from nortl.core.exceptions import AccessAfterReleaseError


def test_memory_map_1() -> None:
    """Test that the Scratch manager class can be initialized correctly."""
    engine = CoreEngine('my_engine')

    zone = engine.scratch_manager.active_zone
    zone.width = 8

    new_scratch_signal = zone.scratchpad[2:3].as_scratch_signal()
    zone.active_view._scratch_signals.append(new_scratch_signal)  # type: ignore[arg-type]

    # We expect that only the bits reserved by the scratch signal are marked as true

    expected = [False] * 8
    expected[2] = True
    expected[3] = True

    assert zone.scratch_map == expected


def test_memory_map_2() -> None:
    """Test that the Scratch manager class can be initialized correctly."""
    engine = CoreEngine('my_engine')

    zone = engine.scratch_manager.active_zone
    zone.width = 8

    new_scratch_signal_1 = zone.scratchpad[2:3].as_scratch_signal()
    zone.active_view._scratch_signals.append(new_scratch_signal_1)  # type: ignore[arg-type]

    new_scratch_signal_2 = zone.scratchpad[6].as_scratch_signal()
    zone.active_view._scratch_signals.append(new_scratch_signal_2)  # type: ignore[arg-type]

    # We expect that only the bits reserved by the scratch signal are marked as true

    expected = [False] * 8
    expected[2] = True
    expected[3] = True
    expected[6] = True

    assert zone.scratch_map == expected

    new_scratch_signal_2.release()

    expected[6] = False

    assert zone.scratch_map == expected


def test_memory_map_3() -> None:
    """Test that the Scratch manager class can be initialized correctly."""
    engine = CoreEngine('my_engine')

    zone = engine.scratch_manager.active_zone
    zone.width = 8

    new_scratch_signal_1 = zone.scratchpad[3:2].as_scratch_signal()
    zone.active_view._scratch_signals.append(new_scratch_signal_1)  # type: ignore[arg-type]

    new_scratch_signal_2 = zone.scratchpad[6].as_scratch_signal()
    zone.active_view._scratch_signals.append(new_scratch_signal_2)  # type: ignore[arg-type]

    # We expect that only the bits reserved by the scratch signal are marked as true

    expected = [False] * 8
    expected[2] = True
    expected[3] = True
    expected[6] = True

    assert zone.scratch_map == expected

    new_scratch_signal_2.release()

    expected[6] = False

    assert zone.scratch_map == expected


def test_alloc() -> None:
    """Test that the alloc method finds the first valid place for a new variable."""
    engine = CoreEngine('my_engine')

    zone = engine.scratch_manager.active_zone

    zone.width = 16

    new_scratch_signal = zone.scratchpad[2:3].as_scratch_signal()
    zone.active_view._scratch_signals.append(new_scratch_signal)  # type: ignore[arg-type]

    new_scratch_signal = zone.scratchpad[6].as_scratch_signal()
    zone.active_view._scratch_signals.append(new_scratch_signal)  # type: ignore[arg-type]

    assert engine.scratch_manager.active_zone.alloc(1) == 0

    assert engine.scratch_manager.active_zone.alloc(8) == 7

    new_scratch_signal = zone.scratchpad[0].as_scratch_signal()
    zone.active_view._scratch_signals.append(new_scratch_signal)  # type: ignore[arg-type]

    assert engine.scratch_manager.active_zone.alloc(1) == 1


def test_write_after_release() -> None:
    engine = CoreEngine('my_engine')
    test = engine.scratch_manager.create_signal(4)

    engine.sync()

    engine.set(test, 1)
    test.release()

    with pytest.raises(Exception):
        engine.set(test, 1)


def test_read_after_release() -> None:
    engine = CoreEngine('my_engine')
    out = engine.define_output('out', 4)
    test = engine.scratch_manager.create_signal(4)

    engine.sync()

    engine.set(test, 1)
    test.release()

    with pytest.raises(Exception):
        engine.set(out, test)


def test_create_and_release_in_context() -> None:
    engine = CoreEngine('my_engine')
    out = engine.define_output('out', 4, 0)

    with Condition(engine, out == 0):
        test = engine.scratch_manager.create_signal(4)
        engine.set(test, 1)
        test.release()

    with pytest.raises(Exception):
        engine.set(out, test)


def test_auto_release_leaving_context() -> None:
    engine = CoreEngine('my_engine')
    out = engine.define_output('out', 4, 0)

    with Condition(engine, out == 0):
        test = engine.scratch_manager.create_signal(4)
        engine.set(test, 1)

    assert test.released


def test_scratch_as_context() -> None:
    engine = CoreEngine('my_engine')

    engine.sync()

    with engine.scratch_manager.create_signal(4) as test:
        pass

    assert test.released


def test_release_in_forked_process() -> None:
    engine = CoreEngine('my_engine')

    engine.sync()

    with Fork(engine, 'f1') as f1:
        test = engine.scratch_manager.create_signal(4)
        print(test.owner.name)
        engine.set(test, 1)
        test.release()
        engine.sync()

    print(f1.running)

    assert not test.released

    f1.cancel()

    assert test.released


def test_parallel_scratch_variables() -> None:
    engine = CoreEngine('my_engine')

    engine.sync()

    with Fork(engine, 'f1') as f1:
        test1 = engine.scratch_manager.create_signal(4)

    with Fork(engine, 'f2') as f2:
        test2 = engine.scratch_manager.create_signal(4)

    assert not test1.released

    f1.cancel()

    assert test1.released
    assert not test2.released  # type:ignore #Weird behavior

    f2.cancel()

    assert test1.released
    assert test2.released


def test_parallel_scratch_access() -> None:
    engine = CoreEngine('my_engine')

    engine.sync()

    with Fork(engine, 'f1') as f1:
        test1 = engine.scratch_manager.create_signal(4)
        engine.set(test1, 1)

    with Fork(engine, 'f2') as f2:
        test2 = engine.scratch_manager.create_signal(4)
        engine.set(test2, 2)

    assert test1.index != test2.index

    f1.cancel()
    f2.cancel()


def test_scratch_signal_disjoint() -> None:
    engine = CoreEngine('my_engine')

    engine.sync()

    a = engine.define_scratch(2)
    b = engine.define_scratch(3)

    engine.sync()
    engine.sync()

    a.release()

    engine.sync()
    engine.sync()

    c = engine.define_scratch(2)
    engine.sync()

    assert not a.states_disjoint(b)
    assert a.states_disjoint(c)


def test_memory_zone() -> None:
    """Test creating a memory zone and allocating signals in it."""
    engine = CoreEngine('my_engine')

    local_signal = engine.define_local('local_signal', 1)
    engine.sync()

    # Create a scratch signal, it will be located in the main zone
    outside_signal = engine.scratch_manager.create_signal(1)
    assert outside_signal.zone == engine.scratch_manager.main_zone

    # The signal can be used as usual
    engine.set(local_signal, outside_signal)
    engine.sync()

    # Create a memory zone
    zone = engine.scratch_manager.create_zone()
    with zone:
        inner_signal = engine.define_scratch(1)
        assert inner_signal.zone is not engine.scratch_manager.main_zone
        assert inner_signal.zone is zone  #  The context manager actually

        # The signal inside the zone can also be used as usual
        engine.set(local_signal, inner_signal)
        engine.sync()

        # The outer signal can still be accessed (the zone is suspended)
        engine.set(local_signal, outside_signal)
        engine.sync()

    # However, after the zone is left, the signals defined within are no longer accessible
    with pytest.raises(AccessAfterReleaseError):
        engine.set(local_signal, inner_signal)


def test_scratch_signal_reclaim() -> None:
    """Tests that scratch signals can be reclaimed.

    This is danger zone!
    """
    engine = CoreEngine('my_engine')
    local_signal = engine.define_local('local_signal', 1)
    engine.sync()

    # Create a memory zone and a scratch signal
    zone = engine.scratch_manager.create_zone()
    with zone as view:  # Save reference to view, normally this is not needed
        inner_signal = engine.define_scratch(1)

    # There can be situations, where signals in a memory zone need to be "revisited",
    # e.g. when copying signals out of a segment. To allow this without disabling access checks, the signal can be reclaimed.

    # Recover the correct view from where inner_signal originates.
    with zone.recover(view):
        # Now, the scratch signal can be reclaimed,
        inner_signal.reclaim()

        engine.set(local_signal, inner_signal)
        engine.sync()

        # Multiple reclaims indicate systematic issues and are not allowed
        with pytest.raises(RuntimeError):
            inner_signal.reclaim()

    # When entering the zone normally, a new view is created. This signal is not accessible.
    with zone:
        # The signal cannot be reclaimed, because it is in the wrong view
        with pytest.raises(RuntimeError):
            inner_signal.reclaim()



"""Test operation framework."""

import pytest

from nortl.core import Const, Signal
from nortl.core.engine import CoreEngine


def test_arithmethics(a: Signal, b: Signal) -> None:
    """Test all arithmetic operations."""
    # Addition
    assert (a + 1).render() == '(a + 1)'  # Note that all two-sided operations are wrapped in brackets
    assert (1 + a).render() == '(1 + a)'

    # Addition short circuits for addition of 0
    assert (a + 0).render() == 'a'
    assert (0 + a).render() == 'a'

    # Substraction
    assert (a - 1).render() == '(a - 1)'
    assert (1 - a).render() == '(1 - a)'

    # Substraction short circuits for substrction of 0
    assert (a - 0).render() == 'a'
    assert (0 - a).render() == '-a'

    # Multiplication
    assert (a * 2).render() == '(a * 2)'
    assert (2 * a).render() == '(2 * a)'

    # Multiplication short circuits for multiplication with 0 or 1
    assert (a * 0).render() == '0'
    assert (0 * a).render() == '0'
    assert (a * 1).render() == 'a'
    assert (1 * a).render() == 'a'

    # Division
    assert (a / 2).render() == '(a / 2)'
    assert (2 / a).render() == '(2 / a)'

    # Division short circuits for division by 0 or 1
    with pytest.raises(ZeroDivisionError):
        a / 0
    assert (0 / a).render() == '0'
    assert (a / 1).render() == 'a'

    # Modulo
    assert (a % 1).render() == r'(a % 1)'
    assert (1 % a).render() == r'(1 % a)'


def test_comparison(a: Signal) -> None:
    """Test all comparison operations.

    There are some details in how comparison behaves, if the left value is not already a Renderable.
    """

    # Equality
    assert (a == 1).render() == '(a == 1)'
    assert (Const(1) == a).render() == '(1 == a)'  # (Non-)Equality only works with Const or Signal on left hand side

    # Unequality
    assert (a != 1).render() == '(a != 1)'
    assert (Const(1) != a).render() == '(1 != a)'

    # Less
    assert (a < 1).render() == '(a < 1)'
    assert (1 < a).render() == '(a > 1)'  # Comparison with integer on left hand side gets reordered by Python
    assert (Const(1) < a).render() == '(1 < a)'  # With a Const, the order stays at is is

    # Less Or Equal
    assert (a <= 1).render() == '(a <= 1)'
    assert (1 <= a).render() == '(a >= 1)'
    assert (Const(1) <= a).render() == '(1 <= a)'

    # Greater
    assert (a > 1).render() == r'(a > 1)'
    assert (1 > a).render() == r'(a < 1)'
    assert (Const(1) > a).render() == '(1 > a)'

    # Greater Or Equal
    assert (a >= 1).render() == '(a >= 1)'
    assert (1 >= a).render() == '(a <= 1)'
    assert (Const(1) >= a).render() == '(1 >= a)'


def test_logic(a: Signal) -> None:
    """Test all logic operations."""
    # And
    assert (a & 1).render() == '(a & 1)'
    assert (1 & a).render() == '(1 & a)'

    # And short-circuits for x & 0
    assert (a & 0).render() == '0'
    assert (0 & a).render() == '0'

    # Or
    assert (a | 1).render() == '(a | 1)'
    assert (1 | a).render() == '(1 | a)'

    # Or short circuits for x | 0
    assert (a | 0).render() == 'a'
    assert (0 | a).render() == 'a'

    # ExclusiveOr
    assert (a ^ 1).render() == '(a ^ 1)'
    assert (1 ^ a).render() == '(1 ^ a)'

    # ExclusiveOr Short Circuits for x ^ 0
    assert (a ^ 0).render() == 'a'
    assert (0 ^ a).render() == 'a'

    # LeftShift
    assert (a << 1).render() == '(a << 1)'
    assert (1 << a).render() == '(1 << a)'

    # LeftShift short circuits for shift of/by 0
    assert (a << 0).render() == 'a'
    assert (0 << a).render() == '0'

    # RightShift
    assert (a >> 1).render() == r'(a >> 1)'
    assert (1 >> a).render() == r'(1 >> a)'

    # RightShift short circuits for shift of/by 0
    assert (a >> 0).render() == 'a'
    assert (0 >> a).render() == '0'


def test_misc(a: Signal) -> None:
    """Test all miscellaneous operations."""
    # Negative
    assert (-a).render() == '-a'
    assert (-Const(1)).render() == '-1'

    # Positive
    assert (+a).render() == 'a'
    assert (+Const(1)).render() == '1'

    # Inversion
    assert (~a).render() == '~(a)'
    assert (~Const(1)).render() == '~(1)'


def test_operation_slice() -> None:
    """Test slicing operations on OperationTraits (except signals)."""
    design_id = Const(0x0815A0, 24)

    # Slices on non-primitive objects must used a mask instead of indexes
    byte0 = design_id[7:0]
    assert byte0.render() == "(24'h0815A0 & 255)"
    assert byte0.operand_width == 8

    byte1 = design_id[15:8]
    assert byte1.render() == "((24'h0815A0 >> 8) & 255)"
    assert byte1.operand_width == 8

    bit = design_id[17]
    assert bit.render() == "((24'h0815A0 >> 17) & 1)"
    assert bit.operand_width == 1


def test_signal_slice(byte: Signal) -> None:
    """Test signal slicing operation."""

    # Single index
    assert byte[0].render() == 'byte[0]'
    assert byte[1].render() == 'byte[1]'

    # Slices are rendered "as-is" - note that this effectively flips the bit order compared to what you would expect.
    assert byte[7:0].render() == 'byte[7:0]'
    assert byte[0:7].render() == 'byte[0:7]'

    # Slices can be used in expressions, like any normal signal would
    assert (2 * byte[7:0] + 1).render() == '((2 * byte[7:0]) + 1)'


def test_nested_signal_slice(engine: CoreEngine, byte: Signal) -> None:
    """Test nested signal slicing operation."""
    # Slices can be nested for convenience, this will slice the base signal (2D signals don't exist)
    assert byte[7:0][0].render() == 'byte[0]'
    assert byte[7:1][2].render() == 'byte[3]'
    assert byte[7:0][7].render() == 'byte[7]'

    # Slices can be nested even more often, it just gets confusing
    assert byte[7:0][6:1][4:1].render() == 'byte[5:2]'
    assert byte[7:0][6:1][4:1][2].render() == 'byte[4]'

    # Nested slices must not go out of range
    with pytest.raises(IndexError):
        byte[7:0][8]
    with pytest.raises(IndexError):
        byte[7:0][-1]

    with pytest.raises(IndexError):
        byte[7:0][8:0]
    with pytest.raises(IndexError):
        byte[7:0][10:9]

    # Nested slices must not revert the bit order of a previous slice
    with pytest.raises(IndexError):
        byte[7:0][0:7]

    # Nested slices even work for signals without fixed width
    WIDTH = engine.define_parameter('WIDTH')  # noqa: N806
    var = engine.define_local('var', WIDTH)
    var[1345:400][100:0]
    # The first slice is always assumed to work, but a nested slice will be validated
    with pytest.raises(IndexError):
        var[1345:0][6789:0]


def test_scratch_signal_slice(scratch_pad: Signal) -> None:
    """Test nested slicing of scratch signals."""
    # Resize scratch pad signal
    scratch_pad.width.update(16)  # type: ignore

    scratch_signal = scratch_pad[15:8].as_scratch_signal()

    scratch_signal[0].render() == 'scratch_signal[8]'
    scratch_signal[7:1].render() == 'scratch_signal[15:9]'
    scratch_signal[7:1][2].render() == 'scratch_signal[11]'


def test_signal_operations(a: Signal, b: Signal, c: Signal) -> None:
    """All operations also work on signals."""
    assert (a + b).render() == '(a + b)'

    # Operations can be chained
    # Wrapping all operations in brackets may be ugly, but ensures that the term order is correctly conveyed.
    # For equally-ranked operations, the expression is simply grouped from left-to-right
    assert (a + b + c).render() == '((a + b) + c)'

    # For non-equally-ranked operations, the implicit term order is used, or explicitely set
    assert (a + b * c).render() == '(a + (b * c))'
    assert ((a + b) * c).render() == '((a + b) * c)'

    # Single-value operations like inversions obviously need no wrapping
    assert (a + b * ~c).render() == '(a + (b * ~(c)))'



from typing import List, Union

from nortl.core.operations import Const
from nortl.core.protocols import OperationProto, Renderable


def extract_operands(obj: Union[Renderable, OperationProto], keep_const: bool = False) -> List[Union[Renderable, OperationProto]]:
    ret = []

    if hasattr(obj, 'operands'):
        for op in obj.operands:
            ret += extract_operands(op, keep_const)
    else:
        if keep_const or not isinstance(obj, Const):
            ret.append(obj)

    return ret



"""Verilog structural elements: ports, declarations, and modules.

This module provides classes for constructing Verilog module declarations,
port declarations, signal declarations, and complete modules.

Example:
    >>> from nortl.renderer.verilog_utils import (
    ...     VerilogModule, VerilogPortDeclaration, VerilogDeclaration
    ... )
    >>>
    >>> # Create a module with ports and signals
    >>> module = VerilogModule("counter")
    >>> module.ports.append(VerilogPortDeclaration("input", "clk"))
    >>> module.ports.append(VerilogPortDeclaration("input", "reset"))
    >>> module.ports.append(VerilogPortDeclaration("output", "count", 4))
    >>> module.signals.append(VerilogDeclaration("reg", "count", 4))
    >>>
    >>> print(module.render())
    module counter (
        input clk,
        input reset,
        output [3:0] count
    );
    reg [3:0] count;
    endmodule
"""

from typing import Dict, List, Optional, Union

from .formatter import VerilogFormatter
from .utils import VerilogRenderable


class VerilogPortDeclaration:
    """Represents a Verilog port declaration.

    Args:
        verilog_type: The port type (e.g., 'input', 'output', 'inout').
        name: The port name or list of port names.
        width: The port width (int for fixed width, str or VerilogRenderable for expression).

    Example:
        >>> port = VerilogPortDeclaration("input", "clk")
        >>> print(port.render())
        input clk
        >>>
        >>> # Port with width
        >>> port_w = VerilogPortDeclaration("output", "data", 8)
        >>> print(port_w.render())
        output [7:0] data
        >>>
        >>> # Multiple ports
        >>> ports = VerilogPortDeclaration("input", ["clk", "reset"])
        >>> print(ports.render())
        input clk, reset
    """

    def __init__(self, verilog_type: str, name: Union[str, List[str]], width: Union[int, str, VerilogRenderable] = 0) -> None:
        """Initialize the port declaration.

        Args:
            verilog_type: The port type.
            name: The port name or list of port names.
            width: The port width.
        """
        self.verilog_type = verilog_type
        self.name = name
        self.width = width

    def render(self) -> str:
        """Render the port declaration.

        Returns:
            The Verilog port declaration string.
        """
        parts: List[str] = []
        parts.append(self.verilog_type)

        if isinstance(self.width, int):
            if self.width > 1:
                parts.append(f'[{self.width - 1}:0]')
        else:
            parts.append(f'[{self.width}-1:0]')

        if isinstance(self.name, list):
            parts.append(', '.join(self.name))
        else:
            parts.append(self.name)
        return ' '.join(parts)


class VerilogDeclaration:
    """Represents a Verilog signal or variable declaration.

    Args:
        verilog_type: The Verilog type (e.g., 'reg', 'wire', 'logic').
        name: The signal name or list of signal names.
        width: The signal width.
        connections: Dictionary of named connections for module instances.
        params: Dictionary of parameters for parameterized declarations.
        members: For enums, either a list of member names or a dict of name->value mappings.

    Example:
        >>> # Simple signal declaration
        >>> signal = VerilogDeclaration("reg", "count", 4)
        >>> print(signal.render())
        reg [3:0] count
        >>>
        >>> # Signal with connections (module instance)
        >>> instance = VerilogDeclaration("counter", "u_counter", None, {
        ...     '.clk(clk)',
        ...     '.reset(reset)'
        ... })
        >>> print(instance.render())
        counter (.clk(clk), .reset(reset))
        >>>
        >>> # Enum declaration
        >>> state_enum = VerilogDeclaration("enum", "state", None, None, None, ["IDLE", "RUNNING", "DONE"])
        >>> print(state_enum.render())
        enum {IDLE, RUNNING, DONE} state
        >>>
        >>> # Enum with values
        >>> state_enum_val = VerilogDeclaration("enum", "state", None, None, None, {"IDLE": 0, "RUNNING": 1, "DONE": 2})
        >>> print(state_enum_val.render())
        enum {IDLE = 0, RUNNING = 1, DONE = 2} state
    """

    def __init__(
        self,
        verilog_type: str,
        name: Union[str, List[str]],
        width: Union[int, str, VerilogRenderable] = 0,
        connections: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, str]] = None,
        members: Optional[Union[List[str], Dict[str, int]]] = None,
    ) -> None:
        """Initialize the declaration.

        Args:
            verilog_type: The Verilog type.
            name: The signal name.
            width: The signal width.
            connections: Named connections for module instances.
            params: Parameters for parameterized declarations.
            members: Enum members.
        """
        self.verilog_type = verilog_type
        self.name = name
        self.connections = connections
        self.params = params
        self.members = members
        self.width = width

    def add_member(self, name: str, value: Optional[Union[int]] = None) -> None:
        """Add an enum member.

        Args:
            name: The member name.
            value: The member value (required for encoded enums).

        Raises:
            RuntimeError: If adding to an enum without proper configuration.
        """
        if self.members is None:
            if value is not None:
                self.members = {}
            else:
                self.members = []

        if isinstance(self.members, list):
            if value is not None:
                raise RuntimeError('Cannot add an encoded member to an enum without all items to be encoded!')
            self.members.append(name)
        else:
            if value is None:
                raise RuntimeError('Cannot add an un-encoded member to an enum where all items are encoded!')
            self.members[name] = value

    def add_parameter(self, name: str, value: Union[str, int]) -> None:
        """Add a parameter.

        Args:
            name: The parameter name.
            value: The parameter value.
        """
        if self.params is None:
            self.params = {}

        if isinstance(value, int):
            self.params[name] = str(value)
        else:
            self.params[name] = value

    def add_connection(self, src: str, tgt: str) -> None:
        """Add a named connection.

        Args:
            src: The source port name.
            tgt: The target signal name.
        """
        if self.connections is None:
            self.connections = {}

        self.connections[src] = tgt

    def render(self) -> str:  # noqa: C901
        """Render the declaration.

        Returns:
            The Verilog declaration string.
        """
        content: List[str] = []
        connection_lst = []
        param_lst = []

        if self.connections is not None:
            connection_lst = [f'.{x}({y})' for x, y in self.connections.items()]
        if self.params is not None:
            param_lst = [f'.{x}({y})' for x, y in self.params.items()]

        name_str = ''

        content.append(self.verilog_type)

        if isinstance(self.name, list):
            name_str = ', '.join(self.name)
        else:
            name_str = self.name

        if isinstance(self.width, int):
            if self.width > 1:
                content.append(f'[{self.width - 1}:0]')
        else:
            content.append(f'[{self.width}-1:0]')

        if self.verilog_type.startswith('enum'):
            if self.members is None:
                raise RuntimeError(f'Tried to create enum {name_str} without any values. Something is wrong here.')

            if isinstance(self.members, List):
                item_str = ', '.join(self.members)
            else:
                item_str = ', '.join([f'{item} = {value}' for item, value in self.members.items()])

            content.append(f'{{{item_str}}}')

        if self.params is not None:
            content.append(f'#({", ".join(param_lst)})')

        content.append(name_str)

        if not any(t in self.verilog_type for t in ['wire', 'logic', 'reg']) and not self.verilog_type.startswith(
            'enum'
        ):  # other net types are out of scope for now
            content.append(f'({", ".join(connection_lst)})')

        return ' '.join(content)


class VerilogModule:
    """Represents a complete Verilog module.

    Args:
        name: The module name.

    Example:
        >>> module = VerilogModule("counter")
        >>> module.ports.append(VerilogPortDeclaration("input", "clk"))
        >>> module.ports.append(VerilogPortDeclaration("input", "reset"))
        >>> module.ports.append(VerilogPortDeclaration("output", "count", 4))
        >>> module.signals.append(VerilogDeclaration("reg", "count", 4))
        >>> print(module.render())
        module counter (
            input clk,
            input reset,
            output [3:0] count
        );
        reg [3:0] count;
        endmodule
        >>>
        >>> # Module with parameters
        >>> param_module = VerilogModule("multiplier")
        >>> param_module.parameters = {"WIDTH": 8}
        >>> param_module.ports.append(VerilogPortDeclaration("input", ["a", "b"]))
        >>> param_module.ports.append(VerilogPortDeclaration("output", "result", "WIDTH"))
        >>> print(param_module.render())
        module multiplier #(
            WIDTH
        ) (
            input a,
            input b,
            output [WIDTH-1:0] result
        );
        endmodule
    """

    def __init__(self, name: str) -> None:
        """Initialize the module.

        Args:
            name: The module name.
        """
        self.name = name
        self.ports: List[VerilogPortDeclaration] = []
        self.parameters: Dict[str, Optional[Union[int, str]]] = {}
        self.signals: List[VerilogDeclaration] = []
        self.instances: List[VerilogDeclaration] = []
        self.functionals: List[VerilogRenderable] = []

    def render(self) -> str:
        """Render the module.

        Returns:
            The complete Verilog module string.
        """
        content = []
        line = f'module {self.name} '

        if len(self.parameters) != 0:
            params = []
            for p, val in self.parameters.items():
                if val is None:
                    params.append(f'parameter {p}')
                else:
                    params.append(f'parameter {p} = {val}')

            line = f'{line} #(\n{",\n".join(params)}) '

        ports = [p.render() for p in self.ports]

        line = f'{line} ({",\n".join(ports)});'

        content.append(line)

        content.extend([item.render() + ';' for item in self.signals + self.instances])
        content.extend([item.render() for item in self.functionals])

        content.append('endmodule')

        code = '\n'.join(content)

        return VerilogFormatter(code).format()

    def get_instance(self, name: str) -> VerilogDeclaration:
        """Get a module instance by name.

        Args:
            name: The instance name.

        Returns:
            The VerilogDeclaration for the instance.

        Raises:
            RuntimeError: If the instance is not found.
        """
        for item in self.instances:
            if item.name == name:
                return item
        raise RuntimeError(f'Could not find an instance named {name}.')



"""Provides type aliases for common types."""

from typing import TYPE_CHECKING, Optional

from typing_extensions import TypeAlias

__all__ = [
    'IntSlice',
]

# slice became generic in 2024, but is somewhat broken (still includes Any?))
if TYPE_CHECKING:
    IntSlice: TypeAlias = slice[Optional[int], Optional[int], Optional[int]]
else:
    IntSlice = slice  # type: ignore[misc]



from typing import Any, Dict

from .debug import DebugEntity


class NamedEntity(DebugEntity):
    """Model a named something in Verilog and implement the feature to add metadata.

    This class will (in future) ensure that there are no naming collisions since all names of verilog objects will be stored here.
    To add traceability, this class inherits from DebugEntity that stores the current execution trace.
    """

    def __init__(self, name: str) -> None:
        super().__init__()
        self._metadata: Dict[str, Any] = {}
        self._name = name

    @property
    def name(self) -> str:
        """Just return name."""
        return self._name

    def get_metadata(self, key: str, default: Any = None) -> Any:
        """Get metadata item."""
        return self._metadata.get(key, default)

    def set_metadata(self, key: str, value: Any) -> None:
        """Set metadata item."""
        self._metadata[key] = value

    def has_metadata(self, key: str) -> bool:
        """Test if a certain key is present in metadata."""
        return key in self._metadata



"""This module will contain helpers for writing sub-engines and wrapping existing functions outside of the core."""

from .condition import Condition, ElseCondition
from .fork_join import Fork
from .loop import ForLoop, WhileLoop
from .segment import Segment

__all__ = [
    'Condition',
    'ElseCondition',
    'ForLoop',
    'Fork',
    'Segment',
    'WhileLoop',
]



"""Describes engine states as Python object."""

from abc import ABCMeta, abstractmethod
from typing import Dict, Final, Iterable, Iterator, List, Literal, Mapping, Optional, Sequence, Set, Tuple, Union

from typing_extensions import Self

from nortl.core.exceptions import ConflictingAssignmentError, ForbiddenAssignmentError, TransitionLockError, TransitionRestrictionError
from nortl.core.operations import All, Any, to_renderable

from .common import NamedEntity
from .protocols import AssignmentTarget, EngineProto, Renderable, RenderableSelector, Selector, StateProto, WorkerProto

__all__ = [
    'State',
]


class BaseAssignment(metaclass=ABCMeta):
    """Baseclass for assignments."""

    def __init__(self, signal: AssignmentTarget) -> None:
        self.signal = signal


class Assignment(BaseAssignment):
    """Regular assignment."""

    unconditional: Final = True

    def __init__(self, signal: AssignmentTarget, value: Renderable) -> None:
        super().__init__(signal)
        self.value = value

    def __format__(self, format_spec: str) -> str:
        return f'{self.signal} = {self.value}'


class _ConditionalAssignmentMixin(metaclass=ABCMeta):
    @property
    @abstractmethod
    def cases(self) -> Sequence[Tuple[Renderable, Union[Renderable, 'SelectorAssignment']]]:
        pass

    def flatten_cases(self) -> Sequence[Tuple[Renderable, Renderable]]:
        """Flatten cases."""
        result: List[Tuple[Renderable, Renderable]] = []

        for condition, value in self.cases:
            if isinstance(value, SelectorAssignment):
                all_conditions = []
                for sub_condition, sub_value in value.flatten_cases():
                    result.append((All(condition, sub_condition), sub_value))
                    all_conditions.append(sub_condition)

                default_condition = All(condition, ~Any(*all_conditions))
                if isinstance(value.default, SelectorAssignment):
                    for sub_condition, sub_value in value.default.flatten_cases():
                        result.append((All(default_condition, sub_condition), sub_value))
                elif value.default is not None:
                    result.append((default_condition, value.default))
            else:
                result.append((condition, value))
        return result


class ConditionalAssignment(_ConditionalAssignmentMixin, BaseAssignment):
    """Assignment creating a conditional assignment."""

    unconditional: Final = False
    priority: Final = False
    default: Final = None

    def __init__(self, signal: AssignmentTarget, value: Renderable, condition: Renderable) -> None:
        super().__init__(signal)
        self.value = value
        self.condition = condition

    @property
    def cases(self) -> Tuple[Tuple[Renderable, Renderable]]:
        return ((self.condition, self.value),)

    def __format__(self, format_spec: str) -> str:
        return f'{self.signal} = {self.value} if {self.condition}'


class SelectorAssignment(_ConditionalAssignmentMixin, BaseAssignment):
    """Assignment creating a selector expression.

    A selector assignment effectively bundles mulitple conditional assignments. It also supports nested cases.
    If a signal has a selector assignment, no other assignments are allowed.
    """

    unconditional: Final = False

    def __init__(self, signal: AssignmentTarget, selector: RenderableSelector, allow_short_circuit: bool = False) -> None:
        super().__init__(signal)
        self._selector = selector
        self._allow_short_circuit = allow_short_circuit
        self._priority: bool
        self._cases: Sequence[Tuple[Renderable, Union[Renderable, 'SelectorAssignment']]]
        self._default: Optional[Union[Renderable, 'SelectorAssignment']]
        self._update_cases()

    @property
    def selector(self) -> RenderableSelector:
        return self._selector

    @selector.setter
    def selector(self, value: RenderableSelector) -> None:
        self._selector = value
        self._update_cases()

    @property
    def priority(self) -> bool:
        """Indicates if the cases of the selector assignment are in prioritized order.

        This is detected based on the presence of a 'default' condition.
        """
        return self._priority

    @property
    def cases(self) -> Sequence[Tuple[Renderable, Union[Renderable, 'SelectorAssignment']]]:
        """Sequence of cases for the selector assignment.

        Each case is represented by a tuple of (condition, value). The value can be a selector assignment of its own.
        """
        return self._cases

    @property
    def default(self) -> Optional[Union[Renderable, 'SelectorAssignment']]:
        """Default value, if no condition is met."""
        return self._default

    def __format__(self, format_spec: str) -> str:
        return f'{self.signal} = {self.selector}'

    def _update_cases(self) -> None:  # noqa: C901
        """Compute cases, default and priority."""

        cases: List[Tuple[Renderable, Union[Renderable, 'SelectorAssignment']]] = []
        default: Optional[Union[Renderable, 'SelectorAssignment']] = None
        priority: bool = False

        # Find the default and copy the rest into a new selector
        selector: Dict['Renderable', Union['Renderable', 'RenderableSelector']] = {}
        for condition, value in self.selector.items():
            if isinstance(condition, str) and condition == 'default':
                # If a default is found, the selector is treated as prioritized
                priority = True
                if isinstance(value, Mapping):
                    default = SelectorAssignment(self.signal, value, allow_short_circuit=self._allow_short_circuit)
                else:
                    default = value
            else:
                selector[condition] = value

        # Process conditions
        for condition, value in selector.items():
            short_circuit = False
            if condition.is_constant:
                # Filter unreachable cases and short circuit: If the current case is always True, the remaining cases are unreachable
                if condition.value == 0:  # type: ignore[attr-defined]
                    continue
                elif condition.value == 1:  # type: ignore[attr-defined]
                    if priority:
                        # Replace the default and stop processing the rest of the cases
                        # This is only done, if the selector was found to have a priority before
                        if isinstance(value, Mapping):
                            default = SelectorAssignment(self.signal, value, allow_short_circuit=self._allow_short_circuit)
                        else:
                            default = value
                        break
                    elif not self._allow_short_circuit:
                        raise RuntimeError(
                            'Selector assignment contains an case that is constantly True.\n'
                            'However, the selector cases were neither in prioritized order, nor was it created with `allow_short_circuit=True`.\n'
                            "Selector assignments are treated as prioritized, if the contain a 'default' case.\n"
                            'To allow noRTL to remove all other cases and only keep the constantly True one, set `allow_short_circuit=True`.'
                        )
                    else:
                        # Short circuit, remove all other cases
                        cases = []
                        short_circuit = True

            if isinstance(value, Mapping):
                cases.append((condition, SelectorAssignment(self.signal, value, allow_short_circuit=self._allow_short_circuit)))
            else:
                cases.append((condition, value))
            if short_circuit:
                break

        self._cases = cases
        self._default = default
        self._priority = priority


def selector_to_renderable(selector: Selector) -> RenderableSelector:
    """Convert conditions and levels of selector to renderables."""
    new_selector: Dict[Union[Renderable, Literal['default']], Union[Renderable, RenderableSelector]] = {}
    for condition, value in selector.items():
        if isinstance(condition, str) and condition == 'default':
            pass
        else:
            condition = to_renderable(condition)

        if isinstance(value, Mapping):
            value = selector_to_renderable(value)
        else:
            value = to_renderable(value)
        new_selector[condition] = value

    return new_selector


AnyAssignment = Union[Assignment, ConditionalAssignment, SelectorAssignment]


class State(NamedEntity):
    """Representation of an engine state.

    An engine state is characterized by a set of assignments to signals and a set of
    conditions, when to enter which next state (transitions).

    Each state belongs to a worker. Transitions can only be created betwen states of the same worker.
    """

    def __init__(self, worker: WorkerProto, name: str, allow_assignments: bool = True):
        """Initialize a state.

        Arguments:
            worker: Engine worker.
            name: State name.
            allow_assignments: If the state allows assignments. This is used for internal purposes.
        """
        super().__init__('')
        self._worker = worker

        # Use setter for validation
        self.name = name

        self._allow_assignments = allow_assignments

        self._assignments: List[AnyAssignment] = []
        self._assigned_signal_names: Set[str] = set()

        # Transition <Condition>, <next value of state variable>
        self._transitions: List[Tuple[Renderable, Self]] = []
        self._restricted_state: Optional[State] = None
        self._transitions_locked = False

        # Store for debug prints. These will not be rendered to synthesizeable constructs
        self._prints: List[Tuple[str, Tuple[Renderable, ...]]] = []
        self._printfs: Dict[str, List[Tuple[str, Tuple[Renderable, ...]]]] = {}

        # Store stack trace at point of creation
        self.engine.tracer.add_metadata(self, 'stack@creation', profile=True)

        # Store all currently active scratch signals for later scratch pad optimization
        self.active_scratch_signals = [
            s for s in self.engine.scratch_manager.active_zone.scratch_signals if not s.released and s.owner == self.engine.current_thread
        ]
        for zone in self.engine.scratch_manager.suspended_zones:
            self.active_scratch_signals.extend([s for s in zone.scratch_signals if not s.released and s.owner == self.engine.current_thread])

    @property
    def engine(self) -> EngineProto:
        """Engine that this state belongs to."""
        return self.worker.engine

    @property
    def worker(self) -> WorkerProto:
        """Engine worker that this state belongs to."""
        return self._worker

    @property
    def name(self) -> str:
        """State name.

        If the worker for this state is not the main worker, the name of the state must be prefixed with the name of the current worker.
        The prefix is automatically added, if missing.
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """State name.

        If the worker for this state is not the main worker, the name of the state must be prefixed with the name of the current worker.
        The prefix is automatically added, if missing.
        """

        # Prefix the worker name if missing. The prefix is omitted for the main worker.
        name = self.worker.create_scoped_name(name)

        if name == self.name:
            return

        if name in self.worker.state_names:
            raise KeyError(f'State {name} already exists.')

        # Update set of state names in worker
        if self.name != '':
            self.worker.state_names.discard(self.name)
        self.worker.state_names.add(name)

        self._name = name

    # Assignment Management
    @property
    def allow_assignments(self) -> bool:
        """If this state allows assignments."""
        return self._allow_assignments

    @property
    def assignments(self) -> Sequence[AnyAssignment]:
        """Sequence of assignments for this state."""
        return self._assignments

    def add_assignment(self, signal: AssignmentTarget, value: Renderable, condition: Optional[Renderable] = None) -> None:
        """Add assignment to this state.

        !!! warning
            Conditional assignments bypass the multi-assignment check for overlap with other conditional assignments!
        """
        if not self.allow_assignments:
            raise ForbiddenAssignmentError(f'State {self.name} does not allow assignments.')

        # Check if signal is already assigned
        for assignment in self.get_assignments(signal):
            # Ignore conditional assignments
            if isinstance(assignment, ConditionalAssignment) and condition is not None:
                continue

            # Test overlap
            overlap = signal.overlaps_with(assignment.signal)

            if overlap == 'partial':
                raise ConflictingAssignmentError(
                    f'State {self.name} already has an assignment to signal {signal.name} that partially overlaps with the new assignment.\n'
                    f'Previous assignment was {assignment}, new assignment is {signal} = {value}.'
                    '\nRefusing to overwrite the signal.'
                )

            # Selector assignments prevent all other assignments
            if overlap is True:
                if isinstance(assignment, SelectorAssignment):
                    raise ConflictingAssignmentError(
                        f'State {self.name} already has an selector assignment to signal {signal.name}.\nIt is not possible to add another assignment.'
                    )

                # Check equality of the signals by rendering them. This could cause false-negatives, but is the easiest way.
                if value.render() != assignment.value.render():
                    raise ConflictingAssignmentError(
                        f'State {self.name} already has an assignment to signal {signal.name}.\n'
                        f'Previous value was {assignment.value}, new value would be {value}. Refusing to overwrite the signal.'
                    )

        # Save new assignment
        if condition is None:
            self._assignments.append(Assignment(signal, value))
        else:
            self._assignments.append(ConditionalAssignment(signal, value, condition))

        self._assigned_signal_names.add(signal.name)

    def add_selector_assignment(self, signal: AssignmentTarget, selector: RenderableSelector, allow_short_circuit: bool = False) -> None:
        """Add selector assignment to this state.

        !!! warning
            Selector assignments bypass the multi-assignment check for partial overlaps!
        """
        if not self.allow_assignments:
            raise ForbiddenAssignmentError(f'State {self.name} does not allow assignments.')

        # Check if signal is already assigned
        for assignment in self.get_assignments(signal):
            overlap = signal.overlaps_with(assignment.signal)
            if overlap is True:
                raise ConflictingAssignmentError(
                    f'State {self.name} already has an assignment to signal {signal.name}.\nIt is not possible to add a selector assignment.'
                )
            elif overlap == 'partial' and assignment.unconditional:
                raise ConflictingAssignmentError(
                    f'State {self.name} already has an unconditional assignment to signal {signal.name}.\nIt is not possible to add a selector assignment.'
                )

        # Save new assignment
        self._assignments.append(SelectorAssignment(signal, selector, allow_short_circuit=allow_short_circuit))
        self._assigned_signal_names.add(signal.name)

    def get_assignments(self, signal: AssignmentTarget) -> Iterator[AnyAssignment]:
        """Iter over current assignment for a signal.

        Arguments:
            signal: The signal to search for.

        Yields:
            The current assignments for this signal.
        """
        # Use a set of the assigned signal names for quick check. In case of a match, search for the signal.
        if signal.name in self._assigned_signal_names:
            for assignment in self.assignments:
                # Check by signal name, will also find slices of the same signal
                if signal.name == assignment.signal.name:
                    yield assignment

    # Transition Management
    @property
    def transitions(self) -> Sequence[Tuple[Renderable, Self]]:
        """List of transitions to other states."""
        return tuple(self._transitions)

    def _add_transition(self, condition: Renderable, state: StateProto) -> None:
        """Add transition to other state."""
        if self._transitions_locked:
            raise TransitionLockError('The transitions for this state have been locked. You cannot add any other transitions.')
        elif self._restricted_state is not None and state is not self._restricted_state:
            raise TransitionRestrictionError(
                f'This state was restricted to transitions to state {self._restricted_state.name}. Unable to add other transitions.'
            )
        if condition.render() != "1'h0":
            self._transitions.append((condition, state))  # type: ignore[arg-type]

    def _restrict_transition(self, state: StateProto) -> None:
        """Restrict transitions to only allow one other state."""
        if self._restricted_state is not None and state is not self._restricted_state:
            raise TransitionRestrictionError(
                f'This state was alreay restricted to transitions to state {self._restricted_state.name}. Unable to restrict for another state.'
            )
        self._restricted_state = state  # type: ignore[assignment]

    def _lock_transitions(self) -> None:
        """Lock the current transitions and prevent any others from being added."""
        self._transitions_locked = True

    # Misc.
    def render(self, target: Optional[str] = None) -> str:
        """Returns the state name for Verilog.

        This function will later contain the logic to transform any string to a correct verilog name.
        """
        return self._name

    def __format__(self, format_spec: str) -> str:
        return self.render()

    def print(self, line: str, *args: Renderable) -> None:
        """Adds a line to the print list that will be processed during simulation."""
        self._prints.append((line, args))

    def printf(self, fname: str, line: str, *args: Renderable) -> None:
        """Store an item that will be output to a file during simulation."""
        if fname in self._printfs:
            self._printfs[fname].append((line, args))
        else:
            self._printfs[fname] = [(line, args)]

    @property
    def signature(self) -> str:
        """Create a string that comprises all assignments and transitions.

        This will be used for redundant state detection.
        """
        ret = ''

        ret += 'Assignments: \n'

        for assignment in sorted(self.assignments, key=lambda x: x.signal.render()):
            if assignment.unconditional:
                ret += f'{assignment.signal} <= {assignment.value}'  # type: ignore[union-attr]
            else:
                ret += '\n'.join(self._extract_conditional_assignment(assignment))  # type: ignore[arg-type]
        ret += '\n'

        ret += 'Transitions: \n'

        for condition, next_state in self.transitions:
            ret += f'if ({condition.render()}), then {next_state.name}\n'

        return ret

    @classmethod
    def _extract_conditional_assignment(cls, assignment: Union[ConditionalAssignment, SelectorAssignment], indent: str = '') -> Iterable[str]:
        """Extract conditional assignment for signature."""
        for i, (condition, value) in enumerate(assignment.cases):
            if i == 0:
                ret = f'{indent}if ({condition}): '
            else:
                ret = f'{indent}elif ({condition}): '

            if hasattr(value, 'cases'):
                ret += f'\n{indent}'
                ret += f'\n{indent}'.join(cls._extract_conditional_assignment(value, indent=indent + '    '))  # type: ignore[arg-type]
            else:
                ret += f'{assignment.signal} <= {value}'
            yield ret



"""Generate the code reference pages."""

from pathlib import Path
from typing import Sequence

import mkdocs_gen_files

INIT_TEMPLATE = """
::: {identifier}
    options:
        members: [{members}]
        members_order: alphabetical
"""


def find_submodules(dir: Path) -> Sequence[str]:
    """Find submodules within a directory."""
    submodules = []
    for path in dir.iterdir():
        if path.is_dir():
            if (path / '__init__.py').exists():
                submodules.append(path.name)
        elif path.suffix == '.py' and path.stem != '__init__':
            submodules.append(path.stem)
    return sorted(submodules)


def gen_ref_pages():
    nav = mkdocs_gen_files.Nav()
    mod_symbol = '<code class="doc-symbol doc-symbol-nav doc-symbol-module"></code>'

    root = Path(__file__).parent.parent.parent
    src = root / 'src'

    for path in sorted(src.rglob('*.py')):
        # Skip Python files that are not in a module
        if not (path.parent / '__init__.py').exists():
            continue

        module_path = path.relative_to(src).with_suffix('')
        doc_path = path.relative_to(src).with_suffix('.md')
        full_doc_path = Path('reference', doc_path)

        parts = tuple(module_path.parts)

        if is_init := parts[-1] == '__init__':
            parts = parts[:-1]
            doc_path = doc_path.with_name('index.md')
            full_doc_path = full_doc_path.with_name('index.md')
        elif parts[-1] == '__main__':
            continue

        nav_parts = [f'{mod_symbol} {part}' for part in parts]
        nav[tuple(nav_parts)] = doc_path.as_posix()

        with mkdocs_gen_files.open(full_doc_path, 'w') as fd:
            identifier = '.'.join(parts)
            if is_init:
                text = INIT_TEMPLATE.format(identifier=identifier, members=','.join(find_submodules(path.parent)))
            else:
                text = f'::: {identifier}\n'
            print(text, file=fd)

        mkdocs_gen_files.set_edit_path(full_doc_path, Path('../../') / path.relative_to(root))

    with mkdocs_gen_files.open('reference/summary.nav', 'w') as nav_file:
        nav_file.writelines(nav.build_literate_nav())


gen_ref_pages()



from contextlib import contextmanager
from types import TracebackType
from typing import Iterator, List, Optional, Sequence, Type

from more_itertools import run_length

from nortl.core.operations import Var
from nortl.core.protocols import EngineProto, MemoryViewProto, PermanentSignal, ScratchSignalProto, SignalProto, ThreadProto
from nortl.core.signal import ScratchSignal, list_indexes


class ScratchManager:
    def __init__(self, engine: EngineProto) -> None:
        self._engine = engine

        # Scratch manager starts out with a main zone
        self._zones: List['MemoryZone'] = []
        self._active_zones: List['MemoryZone'] = []
        self._main_zone = self.create_zone('main')

    @property
    def engine(self) -> EngineProto:
        """NoRTL engine."""
        return self._engine

    # Scratch signal management
    def create_signal(self, width: int) -> ScratchSignal:
        """Create new scratch signal inside the active memory zone."""
        return self.active_zone.create_signal(width)

    # Context management
    def enter_context(self) -> None:
        """Triggers the enter_context function of all managed scratch signals."""
        for scratch_signal in self.active_zone.scratch_signals:
            scratch_signal.enter_context()

    def exit_context(self) -> None:
        """Triggers the exit_context function of all managed scratch signals."""
        for scratch_signal in self.active_zone.scratch_signals:
            scratch_signal.exit_context()

    def force_release_signals_by_thread(self, thread: ThreadProto) -> None:
        """Releases all signals of the current thread. This is triggered by the Thread class after the thread has safely ended."""
        for scratch_signal in self.active_zone.scratch_signals:
            if scratch_signal.owner == thread:
                scratch_signal.release(force=True)

        self.free_accesses_from_thread(thread)

    def free_accesses_from_thread(self, thread: ThreadProto) -> None:
        """Disables all access checks to scratch signals that have their origin in the given thread."""
        for scratch_signal in self.active_zone.scratch_signals:
            scratch_signal.free_access_from_thread(thread)

    # Memory zone managment
    @property
    def zones(self) -> Sequence['MemoryZone']:
        """Memory zones."""
        return self._zones

    @property
    def main_zone(self) -> 'MemoryZone':
        """Main memory zone."""
        return self._main_zone

    @property
    def active_zone(self) -> 'MemoryZone':
        """Active memory zone.

        Zones are activated by entering them as a context manager. Only one zone will be active at a time.
        """
        return self._active_zones[-1]

    @property
    def suspended_zones(self) -> Sequence['MemoryZone']:
        """Suspended memory zones.

        While only one zone will be active in the foreground, the previously active zone(s) will be suspended in the background.
        Signals of suspended zones can still be accessed, but new scratch signals will always be allocated in the active zone.
        """
        return self._active_zones[:-1]

    def create_zone(self, name: Optional[str] = None) -> 'MemoryZone':
        """Create new memory zone.

        Each memory zone consists of individual scratch pad, where scratch signals can be allocated and released.

        The signals inside a zone are automatically released and can no longer be accessed, once it is left.
        However, the allocated scratch pad stays reserved.

        Zones can be re-entered multiple times and will dynamically grow if needed.
        Each time a zone is entered, a new memory view is created. As the memory views are exclusive to each other, they can have different allocations of the scratch signals.

        Once the noRTL engine is completely assembled, zones could be merged, if their active regions don't overlap.
        """
        zone = MemoryZone(self, main_zone=len(self._zones) == 0, name=name)
        self._zones.append(zone)
        return zone


class MemoryZone:
    """Memory zone."""

    def __init__(self, manager: ScratchManager, main_zone: bool = False, name: Optional[str] = None) -> None:
        self._manager = manager
        self._id = len(manager.zones)

        self._width: Optional[Var] = None
        self._scratchpad: Optional[SignalProto] = None

        self._views: List['MemoryView'] = []
        self._active_view: Optional['MemoryView'] = None
        self._main_zone = main_zone

        self._name = f'Zone {self.id}'
        if name is not None:
            self._name += f' ({name})'

        # Main zone always has a single active view
        if main_zone:
            self._create_view()
            self._active_view = self._views[-1]

    @property
    def manager(self) -> ScratchManager:
        """Scratch manager."""
        return self._manager

    @property
    def name(self) -> str:
        """Name of the memory zone."""
        return self._name

    @property
    def main_zone(self) -> bool:
        """Indicates if this zone is the main zone.

        The main zone is always active (or suspended in the background).
        """
        return self._main_zone

    @property
    def id(self) -> int:
        """Numerical identifier of the memory zone."""
        return self._id

    # Scratch pad
    @property
    def width(self) -> int:
        """Width of scratch pad signal."""
        if self._width is None:
            self._create_scratch_pad()
        return self._width.value  # type: ignore[union-attr]

    @width.setter
    def width(self, value: int) -> None:
        """Width of scratch pad signal."""
        if self._width is None:  # pragma: no cover
            self._create_scratch_pad()
        if value < self.width:
            raise ValueError('Scratch pad width must not be decreased.')
        self._width.update(value)  # type: ignore[union-attr]

    @property
    def scratchpad(self) -> SignalProto:
        """Scratch pad."""
        if self._scratchpad is None:  # pragma: no cover
            return self._create_scratch_pad()
        return self._scratchpad

    def _create_scratch_pad(self) -> PermanentSignal:
        """Create scratch pad signal."""
        self._width = Var(0)
        name = 'SCRATCH_SIGNAL'

        if not self.main_zone:
            name += f'_ZONE{self.id}'

        self._scratchpad = self.manager.engine.define_local(name, width=self._width, reset_value=0)
        return self._scratchpad

    @property
    def scratch_map(self) -> Sequence[bool]:
        """Returns a list that shows the state of the currently allocated scratch pad.

        The returned list comprises each bit on the scratch pad as a bool. True-values show, that the bit is currently in use.
        This function is the base for the allocation algorithm that needs to find an appropriate slice of the scratch pad that
        can be used for the next scratch register.

        This function could be realized in a more incremental way in a future version.
        """

        smap = self.width * [False]

        for signal in self.active_view.scratch_signals:
            if not signal.released:
                for index in list_indexes(signal.index):
                    smap[index] = True

        return smap

    def alloc(self, width: int) -> Optional[int]:
        """Looks for a portion of the scratch pad, where a signal with the specified width can be allocated.

        If necessary, the scratch pad is grown to fit the new signal.
        The function returns the starting position for the signal.
        """

        if not isinstance(width, int):
            raise TypeError(f'Width for scratch signal must an integer, got {width}')

        mmap_encoded = run_length.encode(self.scratch_map)

        position = 0
        for used, length in mmap_encoded:
            if not used and length >= width:
                return position
            position = position + length

        # We did not find a position until now, so we extend the scratch pad by enough bits to fit the new signal in.
        self.width += width

        return self.alloc(width)

    # Scratch signal management
    @property
    def scratch_signals(self) -> Sequence['ScratchSignal']:
        """Scratch signals inside the current memory view."""
        return self.active_view.scratch_signals

    def create_signal(self, width: int) -> ScratchSignal:
        """Create new scratch signal inside the memory zone."""
        return self.active_view.create_signal(width)

    # Activation and view management
    @property
    def active(self) -> bool:
        """Indicates if the zone is currently active."""
        return self.manager.active_zone is self

    @property
    def suspended(self) -> bool:
        """Indicates if the zone is currently suspended in the background."""
        return self in self.manager.suspended_zones

    @property
    def active_view(self) -> 'MemoryView':
        """Active view of the zone, containing the allocated scratch signals."""
        if self._active_view is None:
            raise RuntimeError('Memory zone does not have an active view. This is most likely an internal issue.')
        if self.active or self.suspended:
            return self._active_view
        else:
            raise RuntimeError('Memory zone is neither active nor suspended. Its signals cannot be accessed.')

    @property
    def views(self) -> Sequence['MemoryView']:
        """Sequence of memory views."""
        return self._views

    def _create_view(self) -> 'MemoryView':
        self.manager._active_zones.append(self)
        view = MemoryView(self)
        self._views.append(view)
        return view

    def __enter__(self) -> 'MemoryView':
        if self.main_zone:
            raise RuntimeError('This zone is the main zone, unable enter create a new view.')
        view = self._create_view()
        self._active_view = view
        return view

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]) -> None:
        if exc_type is not None:
            return
        if not self.active:
            raise RuntimeError('This zone is not active, unable to deactivate it.')

        if len(self.manager._active_zones) > 1:
            self._release_zone()
            self.manager._active_zones.pop(-1)
        else:
            raise RuntimeError('This zone is the last active one (main zone), unable to deactivate it.')

    @contextmanager
    def recover(self, view: 'MemoryViewProto') -> Iterator[None]:
        """Recover a previous memory view and set it as active.

        This is intended to be used for internal purposes.
        """
        if self.active:
            raise RuntimeError(
                'It is not possible to recover an alternative view while the memory zone is active. This is most likely an internal issue.'
            )

        self.manager._active_zones.append(self)
        self._active_view = view  # type: ignore[assignment]
        yield
        self._release_zone()
        self.manager._active_zones.pop(-1)

    def _release_zone(self) -> None:
        """Release all signals in the zone."""
        for signal in self.active_view.scratch_signals:
            signal.release(force=True)


class MemoryView:
    """Memory view.

    A memory view holds a specific allocation of a memory zone.
    Memory zones can have multiple memory views, that are exclusive to each other.
    """

    def __init__(self, zone: MemoryZone):
        self._zone = zone
        self._scratch_signals: List[ScratchSignal] = []

    @property
    def zone(self) -> MemoryZone:
        """Memory zone."""
        return self._zone

    @property
    def scratch_signals(self) -> Sequence[ScratchSignal]:
        """Sequence of scratch signals allocated in the view."""
        return self._scratch_signals

    def create_signal(self, width: int) -> ScratchSignal:
        """Create new scratch signal inside the memory view."""
        position = self.zone.alloc(width)

        if position is None:
            raise ValueError('Scratch map is full!')

        new_signal = self.zone.scratchpad[position + width - 1 : position].as_scratch_signal(self.zone)
        return new_signal  # type: ignore[return-value]

    def _register_signal(self, signal: ScratchSignalProto) -> None:
        """Register scratch signal.

        This is used for internal purposes only and should never be called manually.

        Scratch signals actively register themselves after creation.
        This ensures, that all scratch signals are tracked correctly.
        """
        self._scratch_signals.append(signal)  # type: ignore[arg-type]



from types import TracebackType

from nortl.core import Any, Const
from nortl.core.protocols import EngineProto, Renderable
from nortl.core.state import State

__all__ = [
    'Condition',
    'ElseCondition',
]


class Condition:
    """Context manager to realize an if-then-else behavior within a noRTL engine."""

    final_state_key: str = 'Condition.start_state'

    def __init__(self, engine: EngineProto, condition: Renderable):
        """Initializes the Condition context manager.

        The state management is as follows:

        *  If we are realizing the first conditional in a state, we need to create a final-state where the engine should be after executing the conditional code.
            * This final-state is stored in the metadata of the current state (and will be reused if another conditional is added)
            * The final-state also has the start state in its meta-data -- if another condition is to be added,
              it should start from the same initial state as the first conditional.
            * After the conditional has been executed, the engine is placed in the final state
        * If we get to realize the second conditional, we recognize that from the metadata of the current state
            * The start and final state are adapted to resemble the behavior of the first conditional

        Finally, this context manager will be made usable with a mixin in the main Engine class using `engine.condition(condition: Renderable)`.

        The Context manager can be used as follows:
        ```python
        from nortl import Engine

        f = Engine('my_engine')
        in_signal = f.define_input('IN')

        # Example usage:
        # <= We start at a start_state
        with engine.condition(in_signal == 1):
            f.sync()
            f.do_something()
            # At the end, the engine goes to an inferred final_state

        with engine.condition(in_signal == 0):
            #The context manager again starts from the same start_state as before
            f.sync()
            f.do_something_else()
            # At the end, the engine goes to an inferred final_state

        # We proceed execution from the final_state

        # When no further conditions are expected, an f.sync() creates a new state and breaks the chain-of-conditionals.

        ```

        Arguments:
            engine: The CoreEngine instance.
            condition: A Renderable signal representing the condition to evaluate.
        """
        self.engine = engine
        self.condition = condition

        if self.engine.current_state.has_metadata(self.final_state_key):
            self.start_state = self.engine.current_state.get_metadata(self.final_state_key)
            self.final_state = self.engine.current_state
            self.engine.current_state = self.start_state
        else:
            self.start_state = self.engine.current_state
            self.final_state = engine.create_state()
            self.final_state.set_metadata(self.final_state_key, self.start_state)

    def __enter__(self) -> None:
        """Executes the provided function if the condition is met."""
        conditional_state = self.engine.create_state()
        self.engine.jump_if(self.condition, conditional_state)
        self.engine.current_state = conditional_state

        # TODO consider moving this into the engine itself
        self.engine.scratch_manager.enter_context()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if exc_type is not None:
            return
        jumpconds = [t[0] for t in self.engine.current_state.transitions]
        if Const(1) not in jumpconds:
            self.engine.jump_if(Const(1, 1), self.final_state)
        self.engine.current_state = self.final_state

        self.engine.scratch_manager.exit_context()


def get_else_condition(state: State) -> Renderable:
    conditions = []
    for c, _ in state.transitions:
        if c.operand_width != 1:
            c = c != 0

        conditions.append(c)

    if len(conditions) == 0:
        return Const(1, 1)

    return ~Any(*conditions)


class ElseCondition(Condition):
    """Context manager to realize an else behavior within a noRTL engine.

    Usage example:

        from nortl import CoreEngine

        f = Engine('my_engine')
        in_signal = f.define_input('IN')

        with engine.condition(in_signal == 1):
            f.sync()
            f.do_something()
        with engine.condition(possible_second_condition):
            f.sync()
            f.do_something()

        with engine.else_condition():
            f.do_something_else()


    """

    def __init__(self, engine: EngineProto):
        state = engine.current_state.get_metadata(self.final_state_key)
        condition = get_else_condition(state)
        super().__init__(engine, condition)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if exc_type is not None:
            return
        super().__exit__(exc_type, exc_val, exc_tb)
        self.engine.sync()



from nortl import Engine


def test_clear_ram(engine: Engine) -> None:
    # Define outputs, reset values are zero
    col_all_on = engine.define_output('col_all_on', width=1)  # Bit width should be optional
    ram_write_zero = engine.define_output('ram_write_zero')
    ram_adr_incr = engine.define_output('ram_adr_incr')

    # Define inputs
    start = engine.define_input('start')
    ram_write_ack = engine.define_input('ram_write_ack')
    ram_end_of_addresses = engine.define_input('ram_end_of_addresses')

    # Define events; could also be done in place
    ram_write_ack_re = ram_write_ack.rising()

    _ = engine.define_local('test_local')

    # Reset State (State 0)
    engine.wait_for(start)

    # Enable all columns (State 1)
    row_loop = engine.current_state  # Remember current state
    engine.set(col_all_on, 1)
    engine.sync()

    # Write Zero to RAM (State 2)
    engine.set_once(ram_write_zero, 1)
    engine.wait_for(ram_write_ack_re == 1)

    # Increment RAM address (State 3)
    engine.set_once(ram_adr_incr, 1)
    engine.sync()

    # Loop (State 4)
    engine.jump_if(ram_end_of_addresses != 1, row_loop, engine.reset_state)

    # set_once merkt für den nächsten Zustand ein Assignment vor
    # es darf nur engine.sync() oder engine.wait_for() benutzt werden, alle anderen Transitions sind verboten (d.h. keine verzweigten States)

    assert len(engine.states[engine.MAIN_WORKER_NAME]) == 5
    assert len(engine.signals) == 9  # 3 inputs, 3 outputs, 2 event outputs (falling is always added too), 1 local signal



from nortl.core.engine import CoreEngine
from nortl.core.protocols import StateProto


class ReachabilityAnalysisMixin(CoreEngine):
    """Reachability analysis by evaluating state transitions.

    This class realizes the functions to execute a simple reachability analysis.
    For this case, we iterate the state graph of the main worker starting from the reset state.
    We visit all states that are reachable in the sense, that there is a transition with a non-zero
    condition can be taken, and mark them as reachable.

    Along this way, the possible thread start states are used to also iterate through the worker's state graphs.

    The information about reachability is stored in the state's metadata in the key 'reachable'
    """

    # Metadata keys to be uses
    REACHABILITY_KEY = 'reachable'

    def reachability_analysis(self) -> None:
        """Entry point for the reachability analysis.

        This function starts a DFS-like iteration of the state graph.
        Initially, all states are marked (via metadata) as unreachable. Then, the DFS will mark all reachable states.
        The iteration startes at the reset state of the main_worker.
        """
        # First, set all states as unreachable and unvisited
        for statelist in self.states.values():
            for state in statelist:
                state.set_metadata(self.REACHABILITY_KEY, False)

        # Mark all reset states as reached
        for worker in self.workers.values():
            worker.reset_state.set_metadata(self.REACHABILITY_KEY, True)

        # Start analysis at the reset state of the main worker
        self._reachability_visit(self.main_worker.reset_state)

    def _reachability_visit(self, state: StateProto) -> None:
        """Visit a state in the reachability analysis.

        This function marks a state as reachable and recurses into the following states.
        The following states are only visited, if the transitions are not constantly off
        (i.e. have a condition being statically False).

        For including the fork-join methods, the recursion also visits the start states of forked processes.
        This is possible, since the states forking new threads are marked via metadata key 'Forked Processes'.
        """
        state.set_metadata(self.REACHABILITY_KEY, True)

        # Visit transition states
        for condition, next_state in state.transitions:
            if (condition == 0).render() != "1'h1":
                # If the condition is not constantly off, we visit the next state
                if not next_state.get_metadata(self.REACHABILITY_KEY):
                    self._reachability_visit(next_state)

        # Visit possibly forked threads
        if state.has_metadata('Forked Processes'):
            for workername, threadid in state.get_metadata('Forked Processes'):
                # Now we have to find the start state in the worker and visit it
                worker = self.workers[workername]
                for workerstate in worker.states:
                    if workerstate.has_metadata('Fork Select ID'):
                        if workerstate.get_metadata('Fork Select ID') == threadid:
                            self._reachability_visit(workerstate)
                            break

    def prune_unreachable_states(self) -> None:
        """Entry point for state pruning.

        Executes a reachability analysis and deletes all states that are not reachable.
        """
        self.reachability_analysis()

        for worker in self.workers.values():
            worker._states = [state for state in worker.states if state.get_metadata(self.REACHABILITY_KEY)]
            worker._state_names = set([state.name for state in worker.states])



from types import TracebackType
from typing import List

from nortl.core.protocols import EngineProto, ThreadProto, WorkerProto

__all__ = [
    'Fork',
]


class Fork:
    """Context manager to realize a fork-join parallelism pattern within a noRTL engine.

    Its intention is to spawn a new thread that runs concurrently with the main thread.
    The fork creates a new worker (if no free worker is available) and a new thread within that worker.

    Example:
    ```python
    engine = Engine("my_engine")
    out = engine.define_output("test_output", width=8)

    with engine.fork("spawned_thread") as spawned_thread:
        # Code that should run in parallel
        engine.set(out, 42)
        engine.sync()

    # After the fork, the main thread continues while the spawned process is running.
    engine.sync()

    # Join the spawned thread to wait for its completion
    spawned_thread.join()
    ```

    The context manager handles the state transitions and worker allocation for the spawned thread.
    It also manages the signal access permissions to ensure proper synchronization between the
    spawning and spawned threads.

    Arguments:
        engine: The CoreEngine instance.
        threadname: Name of the new thread, optional
    """

    def __init__(self, engine: EngineProto, threadname: str):
        """Initializes the Fork context manager.

        Arguments:
            engine: The CoreEngine instance.
            threadname: Name of the new thread, optional
        """
        self.engine = engine
        self.threadname = threadname

        self.checkpoint_worker = self.engine.current_worker
        self.checkpoint_state = self.engine.current_state
        self.assigned_worker: WorkerProto
        self.assigned_thread: ThreadProto

    def _allocate_worker(self) -> None:
        """Allocate a worker and create thread."""
        free_workers: List[WorkerProto] = [w for w in self.engine.workers.values() if not w.working]

        if free_workers == []:
            self.assigned_worker = self.engine.create_worker()
        else:
            self.assigned_worker = free_workers[0]

        self.assigned_thread = self.assigned_worker.create_thread(self.threadname)

    def __enter__(self) -> ThreadProto:
        """Create a new thread and integrate it into next free worker (or create a new worker)."""
        self._allocate_worker()

        # Switch to worker + state
        self.engine.current_worker = self.assigned_worker
        self.engine.current_state = self.assigned_worker.reset_state

        select_id = len(self.assigned_worker.threads)

        # Create start transition from worker idle into fork body
        self.engine.jump_if(self.assigned_worker.start & (self.assigned_worker.select == select_id), self.engine.next_state)
        self.engine.current_state = self.engine.next_state

        self.engine.set(self.assigned_worker.idle, 0)

        for worker in self.engine.workers.values():
            if worker is self.checkpoint_worker:
                spawning_worker = worker

        spawning_thread = spawning_worker.threads[-1]
        spawning_thread._spawned_threads.append(self.assigned_thread)

        # Adjust signal accesses
        self.engine.signal_manager.free_accesses_from_thread(spawning_thread)
        self.engine.scratch_manager.free_accesses_from_thread(spawning_thread)

        # Add metadata
        self.engine.current_state.set_metadata('Fork Select ID', select_id)

        self.engine.sync()

        self.assigned_thread.active = True

        return self.assigned_thread

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if exc_type is not None:
            return
        # Finalize fork body
        self.engine.sync()
        self.assigned_thread.finish()

        # Return to old worker + state
        self.engine.current_worker = self.checkpoint_worker
        self.engine.current_state = self.checkpoint_state

        thread_id = len(self.assigned_worker.threads)

        # Now the thread has been rendered and so we start the Worker!
        self.engine.set(self.assigned_worker.start, 1)
        self.engine.set(self.assigned_worker.select, thread_id)

        # Inform the state via metadata that it actually started a thread
        metadata = (self.assigned_worker.name, thread_id)
        if self.engine.current_state.has_metadata('Forked Processes'):
            old_data = self.engine.current_state.get_metadata('Forked Processes')
            self.engine.current_state.set_metadata('Forked Processes', old_data.append(metadata))
        else:
            self.engine.current_state.set_metadata('Forked Processes', [metadata])

        self.engine.sync()



from nortl import Engine
from nortl.core import CoreEngine
from nortl.core.constructs.loop import WhileLoop
from nortl.core.operations import Const
from nortl.utils.test_wrapper import NoRTLTestBase


class TestTheTestWrapper(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> Engine:
        f = Engine('my_engine')
        f.define_local('testreg', 4, 0)
        return f

    def dut(self, engine: CoreEngine) -> None:
        engine.set(engine.signals['testreg'], 5)
        engine.sync()

    def verify_final_state(self, engine: CoreEngine) -> None:
        self.assertTrue(engine.signals['testreg'] == 5)


class TestTheTestWrapperWithTB(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> Engine:
        f = Engine('my_engine')
        f.define_local('testreg', 4, 0)
        c = f.define_local('channel', 4, 0)
        c.access_checker.disable_check('identical_rw')
        return f

    def dut(self, engine: CoreEngine) -> None:
        for _ in range(10):
            engine.set(engine.signals['testreg'], engine.signals['channel'])
            engine.sync()

    def the_testbench(self, engine: CoreEngine) -> None:
        engine.set(engine.signals['channel'], 5)
        engine.sync()

    def verify_final_state(self, engine: CoreEngine) -> None:
        self.assertTrue(engine.signals['testreg'] == 5)


class TestTheTestWrapperWithTimeout(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> Engine:
        self.expect_timeout = 1

        f = Engine('my_engine')
        f.define_local('testreg', 4, 0)
        c = f.define_local('channel', 4, 0)
        c.access_checker.disable_check('identical_rw')
        return f

    def dut(self, engine: CoreEngine) -> None:
        with WhileLoop(engine, Const(True)):  # endless loop
            engine.sync()

    def the_testbench(self, engine: CoreEngine) -> None:
        pass

    def verify_final_state(self, engine: CoreEngine) -> None:
        # Normally, we get a fail due to the timeout signal, so we need to decrement the error_ctr to mask ths
        self.assertTrue(engine.signals['timeout'])
        self.finish_simulation()



"""Utility types and helpers for Verilog rendering.

This module provides the VerilogRenderable protocol and helper functions
for converting values to Verilog renderable objects.

Example:
    >>> from nortl.renderer.verilog_utils.utils import VerilogRenderable, to_verilog_renderable
    >>>
    >>> # String values are automatically converted
    >>> renderable = to_verilog_renderable("clk")
    >>> print(renderable.render())
    clk
    >>>
    >>> # VerilogRenderable objects are passed through
    >>> class MyRenderable(VerilogRenderable):
    ...     def render(self) -> str:
    ...         return "my_signal"
    >>>
    >>> renderable2 = to_verilog_renderable(MyRenderable())
    >>> print(renderable2.render())
    my_signal
"""

from typing import Protocol, Union

from nortl.core.operations import RawText


class VerilogRenderable(Protocol):
    """Protocol for Verilog Renderable.

    Compared to the [nortl.core.Renderable][nortl.core.operations.Renderable], it doesn't require the
    [OperationTrait Mixin][nortl.core.operations.OperationTrait] and `target` argument for `render()`.

    Example:
        >>> class SimpleSignal(VerilogRenderable):
        ...     def render(self) -> str:
        ...         return "simple_signal"
        >>>
        >>> signal = SimpleSignal()
        >>> print(signal.render())
        simple_signal
    """

    def render(self) -> str: ...


def to_verilog_renderable(value: Union[str, VerilogRenderable]) -> VerilogRenderable:
    """Convert value to Verilog Renderable.

    Compared to [nortl.core.to_renderable][nortl.core.operations.to_renderable], this also supports strings.

    Args:
        value: The value to convert. Can be a string or a VerilogRenderable.

    Returns:
        A VerilogRenderable instance.

    Example:
        >>> # String input
        >>> renderable = to_verilog_renderable("clk")
        >>> print(renderable.render())
        clk
        >>>
        >>> # VerilogRenderable input
        >>> class MyRenderable(VerilogRenderable):
        ...     def render(self) -> str:
        ...         return "my_signal"
        >>>
        >>> result = to_verilog_renderable(MyRenderable())
        >>> print(result.render())
        my_signal
    """
    if isinstance(value, str):
        return RawText(value)
    else:
        return value



"""Contant value."""

from typing import Optional

from nortl.core.renderers.operations.base import SliceRenderer

__all__ = [
    'Slice',
]


class Slice(SliceRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        if self.value.is_primitive:
            if isinstance(self.index, int):
                return f'{self.value}[{self.index}]'
            else:
                return f'{self.value}[{self.index.start}:{self.index.stop}]'
        else:
            if isinstance(self.index, int):
                mask = 1
                offset = self.index
            else:
                mask = 2 ** (max(self.index.start, self.index.stop) - min(self.index.start, self.index.stop) + 1) - 1
                offset = self.index.stop

            if offset != 0:
                return f'(({self.value} >> {offset}) & {mask})'
            else:
                return f'({self.value} & {mask})'



import math
from typing import Any, Dict, List, Optional, Sequence, Set

from typing_extensions import Self

from nortl.core.exceptions import OwnershipError, UnfinishedForwardDeclarationError

from .common import NamedEntity
from .modifiers import Volatile
from .operations import Const, Var
from .protocols import EngineProto, Renderable, SignalProto, StateProto, WorkerProto
from .state import State


class Worker(NamedEntity):
    """Worker for noRTL engine.

    Each worker represents a state machine, that runs independently from the others.
    The workers store their states. Creating state transitions between different workers is not possible.

    The noRTL engine contains one main worker by default. Additional workers can be created manually.
    The main worker is responsible for managing the reset state for all signals; the other workers do not contain any assignments in the reset state.

    Each worker must have one or more threads.
    """

    def __init__(self, engine: EngineProto, name: str, reset_state_name: str = 'IDLE') -> None:
        """Initialize a new Worker.

        Arguments:
            engine: noRTL engine.
            name: Name of the worker.
            reset_state_name: Name of the reset state for this worker.

        !!! warning

            Workers are not meant to be instantiated manually. Use the [`CoreEngine.create_worker()`][nortl.core.engine.CoreEngine.create_worker] method instead.
        """
        super().__init__(name)

        self._engine = engine
        self._is_main_worker = name == engine.MAIN_WORKER_NAME

        # State tracking
        self._states: List[State] = []
        self._state_names: Set[str] = set()

        # Thread tracking
        self._threads: List['Thread'] = []  # Threads mapped to this worker

        # Create reset state and set to current state
        self._reset_state = self.create_state(reset_state_name, allow_assignments=self.is_main_worker)
        self._current_state = self.reset_state
        self._next_state: Optional[State] = None

        # Internal control signals
        # The main worker does not use these signals
        self._reset_signal: Optional[Volatile[SignalProto]] = None
        self._start_signal: Optional[Volatile[SignalProto]] = None
        self._select_signal: Optional[Volatile[SignalProto]] = None
        self._idle_signal: Optional[Volatile[SignalProto]] = None
        self._select_signal_width: Optional[Var] = None

        # Synchronous reset, inactive by default
        self._sync_reset: Optional[Renderable] = None

    @property
    def engine(self) -> EngineProto:
        """NoRTL engine for this worker."""
        return self._engine

    @property
    def is_main_worker(self) -> bool:
        """If this worker is the main worker of the noRTL engine."""
        return self._is_main_worker

    def create_scoped_name(self, name: str) -> str:
        """Create scoped name for signals or other things.

        If this worker is not the main worker, the name is prefixed with the name of the worker.
        """
        # FIXME decide if worker prefix is omitted for main worker
        if not self.is_main_worker and not name.startswith(self.name):
            name = f'{self.name}_{name}'
        return name

    @property
    def sync_reset(self) -> Renderable:
        """Synchronous reset for worker."""
        if self._sync_reset is None:
            return Const(0)
        return self._sync_reset

    @sync_reset.setter
    def sync_reset(self, value: Renderable) -> None:
        """Synchronous reset for worker."""
        if self._sync_reset is not None:
            raise RuntimeError(f'Synchronous reset for worker {self.name} was already set to {self.sync_reset}.')
        self._sync_reset = value

    # Control Signals for Threads
    # FIXME technically, these signals are only required for Fork/Join management. They serve no purpose for raw Worker + Thread
    @property
    def reset(self) -> SignalProto:
        """Control signal to reset worker.

        If the worker control signal is used, it is automatically added to the synchronous reset.
        """
        if self.is_main_worker:
            raise RuntimeError('Main worker does not use reset signal.')
        elif self._reset_signal is None:
            self._initialize_control_signals()
        return self._reset_signal  # type: ignore[return-value]

    @property
    def start(self) -> SignalProto:
        """Control signal to start worker."""
        if self.is_main_worker:
            raise RuntimeError('Main worker does not use start signal.')
        elif self._start_signal is None:
            self._initialize_control_signals()
        return self._start_signal  # type: ignore[return-value]

    @property
    def select(self) -> SignalProto:
        """Control signal that selects a thread."""
        if self.is_main_worker:
            raise RuntimeError('Main worker does not use select signal.')
        elif self._select_signal is None:
            self._initialize_control_signals()
        return self._select_signal  # type: ignore[return-value]

    @property
    def idle(self) -> SignalProto:
        """Status signal that indicates if the worker is in it's idle state."""
        if self.is_main_worker:
            raise RuntimeError('Main worker does not use idle signal.')
        elif self._idle_signal is None:
            self._initialize_control_signals()
        return self._idle_signal  # type: ignore[return-value]

    def _initialize_control_signals(self) -> None:
        """Initializes built-in signals for the worker."""

        # Local Signals
        self._reset_signal = Volatile(
            self.engine.define_local(self.create_scoped_name('reset'), width=1, reset_value=0, pulsing=True), 'identical_rw', 'exclusive_write'
        )
        self._start_signal = Volatile(
            self.engine.define_local(self.create_scoped_name('start'), width=1, reset_value=0, pulsing=True), 'identical_rw', 'exclusive_write'
        )
        self._select_signal_width = Var(self._get_select_width())
        self._select_signal = Volatile(
            self.engine.define_local(self.create_scoped_name('select'), width=self._select_signal_width, reset_value=0),
            'identical_rw',
            'exclusive_write',
        )

        # the following signal has to be set to zero once a thread is started:
        # FIXME: Set signal to zero and to one after first state
        self._idle_signal = Volatile(self.engine.define_local(self.create_scoped_name('idle'), reset_value=1), 'identical_rw', 'exclusive_write')

        # Tie synchrounous reset to the new reset signal
        self.sync_reset = self.reset

    def _get_select_width(self) -> int:
        """Calculate necessary width for thread select signal."""
        return math.ceil(math.log2(len(self.threads)) + 1)

    def _resize_select_signal(self) -> None:
        """Updates the internal signal width of the select variable."""
        if self._select_signal_width is not None:
            self._select_signal_width.update(self._get_select_width())

    # State management
    @property
    def states(self) -> Sequence[State]:
        """List of states for this worker."""
        return self._states

    @property
    def state_names(self) -> Set[str]:
        """Set of the names of all states for this worker."""
        return self._state_names

    def create_state(self, name: Optional[str] = None, allow_assignments: bool = True, metadata: Dict[str, Any] = {}) -> State:
        """Create a state.

        Arguments:
            name: Optional state name. If no name is provided, it defaults to '<worker_name>_STATE_<id>', where 'id' is the current number of states
                for the worker and 'worker_name' is the name of this worker.

                If this worker is not the main worker, the name of the state must be prefixed with the name of the current worker.
                The prefix is automatically added, if missing.
            allow_assignments: If the state allows assignments. This is used for internal purposes.
            metadata: The metadata that will be added to the newly created state. If not given, the engine's state_metadata_template will be used.

        Returns:
            The created state.
        """
        # Generate default state name
        if name is None:
            name = f'STATE_{len(self.states)}'

        if len(self.states) > 0 and len(self.threads) == 0:
            raise RuntimeError('Worker has no thread, unable to create new states.')

        # Use engine's metadata template if no metadata is provided
        if metadata == {}:
            metadata = self.engine.state_metadata_template

        # State will validate the name
        state = State(self, name, allow_assignments=allow_assignments)

        # Store given metadata in state
        for k, v in metadata.items():
            state.set_metadata(k, v)

        self._states.append(state)
        return state

    @property
    def current_state(self) -> State:
        """Current state.

        Returns:
            The current state.
        """
        return self._current_state

    @current_state.setter
    def current_state(self, state: StateProto) -> None:
        """Current state.

        Arguments:
            state: The new current state.

        Raises:
            UnfinishedForwardDeclarationError: If the next state has been forward-declared and is different from the new current state.
        """
        if state.engine is not self.engine:
            raise OwnershipError('State does not belong to this engine.')
        if state not in self.states:
            raise OwnershipError('State does not belong to this worker.')
        if self._next_state is not None and state is not self._next_state:
            raise UnfinishedForwardDeclarationError(
                'You have forward-declared the next state (by using next_state), but are now trying to set another state as the current one. '
                'This may result in dead-end states and is therefore forbidden. Please switch to the next_state and modify it first.'
            )
        if state is not self.current_state:
            self._next_state = None  # Clear next state
        self._current_state = state  # type: ignore[assignment]

    @property
    def next_state(self) -> State:
        """Forward-declared next state.

        This simplifies the creation of new states for non-branching sections of the state graph (e.g. via sync() or wait_for()).
        When you use next_state you must set it as current_state, before you can set any other state.

        Returns:
            The forward-declared next state.
        """
        if self._next_state is None:
            self._next_state = self.create_state()

        return self._next_state

    @property
    def reset_state(self) -> State:
        """The reset state of the engine.

        This is the initial state from which the engine will start.

        Returns:
            The reset state.
        """
        return self._reset_state

    def leave_foreground(self) -> None:
        """This method must be called when the current worker of the engine is changed.

        Checks if the any forward-declared next-state pending.
        """
        if self._next_state is not None:
            raise UnfinishedForwardDeclarationError(
                'You have forward-declared the next state for the current worker (by using next_state), but are now trying to set another worker as the current one. '
                'This may result in dead-end states and is therefore forbidden. Please finish the current worker first, by switching to the next_state and modifying it.'
            )

    # Thread Management
    @property
    def threads(self) -> Sequence['Thread']:
        """Stack of worker threads."""
        return self._threads

    def create_thread(self, name: Optional[str] = None) -> 'Thread':
        """Create a new thread.

        Arguments:
            name: Name of the thread.

        Returns:
            The created thread.
        """
        # Generate default name
        if name is None:
            if len(self.threads) == 0:
                name = self.engine.MAIN_THREAD_NAME
            else:
                name = f'thread_{len(self.threads)}'

        thread = Thread(self, name)
        self._threads.append(thread)

        if not self.is_main_worker:
            self._resize_select_signal()  # Adjust width of select signal
        return thread

    @property
    def current_thread(self) -> 'Thread':
        """Current thread for this worker."""
        if len(self.threads) == 0:
            raise RuntimeError('Worker has no threads.')
        return self.threads[-1]

    @property
    def working(self) -> bool:
        """If the worker is working."""
        return any([t.running for t in self.threads])

    # Misc.
    # FIXME these helper methods are only used by Fork/Join
    def raise_reset(self) -> None:
        self.engine.set(self.reset, 1)

    def clear_reset(self) -> None:
        self.engine.set(self.reset, 0)
        self.engine.set(self.idle, 1)


class Thread(NamedEntity):
    """Thread for noRTL engine.

    Threads are used to control signal ownership.
    They prevent simultaneous access to signals from two workers, that would cause wrong behavior.
    """

    def __init__(self, worker: WorkerProto, name: str) -> None:
        """Initialize a new thread.

        Arguments:
            worker: Worker for this thread.
            name: Name of the thread.

        !!! warning

            Threads are not meant to be instantiated manually. Use the [`Worker.create_thread()`][nortl.core.process.Worker.create_thread] method instead.
        """
        super().__init__(name)

        self._worker: WorkerProto = worker
        self._is_main_thread = worker.is_main_worker
        self._active: bool = self.is_main_thread  # Main thread is always active

        self._spawned_threads: List[Self] = []

        # Link back the parent thread
        self.parent_thread: Optional[Self] = None
        if name != worker.engine.MAIN_THREAD_NAME:
            self.parent_thread = self.worker.engine.current_thread  # type: ignore[assignment]

    @property
    def engine(self) -> EngineProto:
        """NoRTL engine for this thread."""
        return self.worker.engine

    @property
    def worker(self) -> WorkerProto:
        """Engine worker."""
        return self._worker

    @property
    def is_main_thread(self) -> bool:
        """If this thread is the main thread of the noRTL engine."""
        return self.worker.is_main_worker

    @property
    def active(self) -> bool:
        """If this thread is currently active."""
        # FIXME decide if activity can be tracked automatically
        return self._active

    @active.setter
    def active(self, value: bool) -> None:
        """If this thread is currently active."""
        if self.is_main_thread:
            raise RuntimeError('Main thread cannot be deactivated.')
        self._active = value

    def join(self) -> None:
        """Join a thread.

        This waits for the thread to be finished.
        """
        self.engine.wait_for(self.finished)
        self.active = False
        self.engine.scratch_manager.force_release_signals_by_thread(self)
        self.engine.signal_manager.free_accesses_from_thread(self)

    @property
    def finished(self) -> Renderable:
        return self.worker.idle

    @property
    def running(self) -> bool:
        if self.parent_thread == self.engine.current_thread:
            return self.active

        if self.active:
            return True

        global_call_stack = self.engine.current_thread.call_stack

        for t in self.call_stack:
            if t not in global_call_stack:
                if t.active:
                    return True

        return False

    def cancel(self) -> None:
        """This method sends a synchronous reset to the Thread's worker and recursively to all threads, that are spawned from this thread."""
        for subthread in self._spawned_threads:
            subthread.cancel()

        self.worker.raise_reset()
        self.engine.sync()
        self.worker.clear_reset()
        self.engine.sync()

        self.active = False
        self.engine.scratch_manager.force_release_signals_by_thread(self)
        self.engine.signal_manager.free_accesses_from_thread(self)

    def finish(self) -> None:
        """Finishes a Thread."""
        self.engine.jump_if(Const(1), self.worker.reset_state)
        self.engine.set(self.worker.idle, 1)

    @property
    def call_stack(self) -> List[Self]:
        ret = []

        thread = self

        while thread.parent_thread is not None:
            ret.append(thread.parent_thread)
            thread = thread.parent_thread

        return ret



"""Tests for the ManagementMixin context managers in nortl/__init__.py."""

from nortl import Engine


class TestContextManager:
    """Tests for the context() context manager."""

    def test_context_basic_usage(self) -> None:
        """Test that context() context manager properly scopes scratch variables."""
        engine = Engine('test_engine')
        engine.sync()

        with engine.context():
            test = engine.define_scratch(4)
            engine.set(test, 1)
            engine.sync()

        # After exiting context, scratch signal should be released
        assert test.released

    def test_context_multiple_scopes(self) -> None:
        """Test that multiple context scopes work correctly."""
        engine = Engine('test_engine')
        engine.sync()

        with engine.context():
            test1 = engine.define_scratch(4)
            engine.set(test1, 1)
            engine.sync()

        with engine.context():
            test2 = engine.define_scratch(3)
            engine.set(test2, 2)
            engine.sync()

        assert test1.released
        assert test2.released

    def test_context_nesting(self) -> None:
        """Test that nested context managers work correctly."""
        engine = Engine('test_engine')
        engine.sync()

        with engine.context():
            with engine.context():
                test = engine.define_scratch(4)
                engine.set(test, 1)
                engine.sync()
                assert not test.released

            assert test.released


class TestCollapseSyncManager:
    """Tests for the collapse_sync() context manager."""

    def test_collapse_sync_basic_usage(self) -> None:
        """Test that collapse_sync() marks states as collapsable."""
        engine = Engine('test_engine')

        with engine.collapse_sync():
            # Should be collapsable
            assert engine.state_metadata_template['collapsable'] is True

        # After exiting, should still be collapsable
        assert engine.state_metadata_template['collapsable'] is False

    def test_collapse_sync_multiple_calls(self) -> None:
        """Test that collapse_sync can be called multiple times."""
        engine = Engine('test_engine')

        with engine.collapse_sync():
            assert engine.state_metadata_template['collapsable'] is True

        with engine.collapse_sync():
            assert engine.state_metadata_template['collapsable'] is True

        assert engine.state_metadata_template['collapsable'] is False

    def test_collapse_sync_nesting(self) -> None:
        """Test that nested collapse_sync managers work correctly."""
        engine = Engine('test_engine')

        with engine.collapse_sync():
            with engine.collapse_sync():
                assert engine.state_metadata_template['collapsable'] is True

        assert engine.state_metadata_template['collapsable'] is False

    def test_collapse_sync_nesting_with_multiple_calls(self) -> None:
        """Test that nested collapse_sync managers work correctly."""
        engine = Engine('test_engine')

        with engine.collapse_sync():
            with engine.collapse_sync():
                assert engine.state_metadata_template['collapsable'] is True

            # The outer context manager has priority
            assert engine.state_metadata_template['collapsable'] is True

            with engine.collapse_sync():
                assert engine.state_metadata_template['collapsable'] is True

        assert engine.state_metadata_template['collapsable'] is False



from nortl.core.protocols import ThreadProto

from .debug import DebugEntity


class StaticAccess(DebugEntity):
    """Model for an access to a signal.

    This class is used as a container to hold references to the thread that caused the access and information about what caused the access where in the code.
    The signal object can then hold a list (or deque) of accesses for executing checks for e.g. two parallel-running threads accessing the register.

    Note that this does not include checks like access counters that verify,
    that the number of reads is equal to the number of writes so that no data gets lost.
    This structure is only used for detecting concurrent accesses at assemble-time of the noRTL engine (where the code assembles the internal noRTL engine model).
    Realizing access counters is a task to be performed at other levels of abstraction, e.g. in the rendered Verilog code where also formal properties
    can be included for verification in simulation.
    """

    def __init__(self, thread: ThreadProto) -> None:
        super().__init__()
        self.thread = thread
        self.active = True

    def disable(self) -> None:
        self.active = False



"""Empty state removal optimization for noRTL engines.

This module provides an optimization that removes states from the state graph that have no assignments
and only one transition to a next state with a condition of "1'h1" (always on).
"""

from typing import List, Tuple

from nortl.core.engine import CoreEngine
from nortl.core.protocols import Renderable, StateProto


class EmptyStateRemovalMixin(CoreEngine):
    """Mixin providing empty state removal optimization.

    This optimization removes states that:
    1. Have no assignments
    2. Have exactly one transition
    3. The transition condition is "1'h1" (always true)
    4. Have the metadata 'collapsable' set to True

    Example:
        Consider the state sequence S1 -> S_empty -> S3 where S_empty is a state without assignments
        and with only one transition to S3 with a condition being 1'h1 (always on).
        The optimizer will remove S_empty and wire the state graph to S1 -> S3.
    """

    def _is_empty_state(self, state: StateProto) -> bool:
        """Check if a state has no assignments.

        Arguments:
            state: The state to check.

        Returns:
            True if the state has no assignments, False otherwise.
        """
        return len(state.assignments) == 0

    def _is_single_transition(self, state: StateProto) -> bool:
        """Check if a state has exactly one transition.

        Arguments:
            state: The state to check.

        Returns:
            True if the state has exactly one transition, False otherwise.
        """
        return len(state.transitions) == 1

    def _is_always_true_transition(self, state: StateProto) -> bool:
        """Check if a state's transition condition is "1'h1" (always true).

        Arguments:
            state: The state to check.

        Returns:
            True if the transition condition is "1'h1", False otherwise.
        """
        if not self._is_single_transition(state):
            return False

        condition, _ = state.transitions[0]
        return condition.render() == "1'h1"

    def _is_collapsable(self, state: StateProto) -> bool:
        """Check if a state has the 'collapsable' metadata set to True.

        Arguments:
            state: The state to check.

        Returns:
            True if the state has 'collapsable' metadata set to True, False otherwise.
        """
        return state.has_metadata('collapsable') and state.get_metadata('collapsable') is True

    def _is_removable_state(self, state: StateProto) -> bool:
        """Check if a state meets all criteria for removal.

        A state is removable if:
        1. It has no assignments
        2. It has exactly one transition
        3. The transition condition is "1'h1" (always true)
        4. It has the 'collapsable' metadata set to True

        Arguments:
            state: The state to check.

        Returns:
            True if the state is removable, False otherwise.
        """
        return (
            self._is_empty_state(state)
            and self._is_single_transition(state)
            and self._is_always_true_transition(state)
            and self._is_collapsable(state)
        )

    def _get_predecessor_states(self, state: StateProto) -> List[StateProto]:
        """Get all states that have a transition to the given state.

        Arguments:
            state: The state to find predecessors for.

        Returns:
            List of predecessor states.
        """
        predecessors: List[StateProto] = []
        for worker_states in self.states.values():
            for predecessor in worker_states:
                for condition, next_state in predecessor.transitions:
                    if next_state == state:
                        predecessors.append(predecessor)
                        break
        return predecessors

    def _get_successor_state(self, state: StateProto) -> StateProto | None:
        """Get the successor state of a state.

        Arguments:
            state: The state to get the successor for.

        Returns:
            The successor state if there is exactly one transition, None otherwise.
        """
        if not self._is_single_transition(state):
            return None

        _, successor = state.transitions[0]
        return successor

    def _remove_empty_state(self, state: StateProto) -> None:
        """Remove an empty state and update transitions.

        This function removes the given state and updates all transitions that pointed to it
        to point to its successor state instead.

        Arguments:
            state: The state to remove.
        """
        successor = self._get_successor_state(state)
        if successor is None:
            return

        # Get all predecessor states
        predecessors = self._get_predecessor_states(state)

        # Update transitions in predecessor states
        for predecessor in predecessors:
            new_transitions: List[Tuple[Renderable, StateProto]] = []
            for condition, next_state in predecessor.transitions:
                if next_state == state:
                    new_transitions.append((condition, successor))
                else:
                    new_transitions.append((condition, next_state))
            predecessor._transitions = new_transitions  # type: ignore

        # Remove the state from its worker
        worker = state.worker
        worker._states = [s for s in worker.states if s.name != state.name]  # type: ignore
        worker._state_names.discard(state.name)  # type: ignore

    def empty_state_removal(self) -> None:
        """Remove empty states from the state graph.

        This function iterates through all states and removes those that meet the criteria:
        1. No assignments
        2. Exactly one transition
        3. Transition condition is "1'h1" (always true)
        4. Has 'collapsable' metadata set to True

        The removal is done in a single pass, as removing one state doesn't affect the
        removability of other states (since we only remove states with no assignments).
        """
        for worker_states in self.states.values():
            # Create a list of states to remove (to avoid modifying list while iterating)
            states_to_remove: List[StateProto] = [state for state in worker_states if self._is_removable_state(state)]

            for state in states_to_remove:
                self._remove_empty_state(state)



import networkx as nx
import pytest

from nortl.core.engine import CoreEngine
from nortl.core.operations import Const
from nortl.renderer.networkx_renderer import NetworkXRenderer


def test_networkx_renderer_setup() -> None:
    """Test that the NetworkXRenderer initializes correctly."""
    engine = CoreEngine('test_engine')
    renderer = NetworkXRenderer(engine)
    assert isinstance(renderer, NetworkXRenderer)


def test_networkx_renderer_render_empty_engine() -> None:
    """Test that rendering an empty engine returns an graph with only a reset state."""
    engine = CoreEngine('empty_engine')
    renderer = NetworkXRenderer(engine)
    graph = renderer.render()
    assert isinstance(graph, nx.DiGraph)
    assert len(graph.nodes) == 1
    assert len(graph.edges) == 0


def test_networkx_renderer_render_single_state_engine() -> None:
    """Test that rendering an engine with a single state returns a graph with a single node."""
    engine = CoreEngine('single_state_engine')
    state = engine.create_state('state')  # noqa: F841
    renderer = NetworkXRenderer(engine)
    graph = renderer.render()
    assert isinstance(graph, nx.DiGraph)
    assert len(graph.nodes) == 2
    assert len(graph.edges) == 0


def test_networkx_renderer_render_multiple_states_engine() -> None:
    """Test that rendering an engine with multiple states returns a graph with multiple nodes."""
    engine = CoreEngine('multiple_states_engine')
    state1 = engine.create_state('state1')
    state2 = engine.create_state('state2')
    state1._add_transition(Const(True), state2)
    renderer = NetworkXRenderer(engine)
    graph = renderer.render()
    assert isinstance(graph, nx.DiGraph)
    assert len(graph.nodes) == 3
    assert len(graph.edges) == 1


def test_networkx_renderer_render_engine_with_transitions() -> None:
    """Test that rendering an engine with transitions returns a graph with edges."""
    engine = CoreEngine('engine_with_transitions')
    state1 = engine.create_state('state1')
    state2 = engine.create_state('state2')
    state1._add_transition(Const(True), state2)
    renderer = NetworkXRenderer(engine)
    graph = renderer.render()
    assert isinstance(graph, nx.DiGraph)
    assert len(graph.nodes) == 3
    assert len(graph.edges) == 1
    assert ('state1', 'state2') in graph.edges


def test_networkx_renderer_render_engine_with_multiple_transitions() -> None:
    """Test that rendering an engine with multiple transitions returns a graph with multiple edges."""
    engine = CoreEngine('engine_with_multiple_transitions')
    state1 = engine.create_state('state1')
    state2 = engine.create_state('state2')
    state3 = engine.create_state('state3')
    state1._add_transition(Const(True), state2)
    state1._add_transition(Const(True), state3)
    renderer = NetworkXRenderer(engine)
    graph = renderer.render()
    assert isinstance(graph, nx.DiGraph)
    assert len(graph.nodes) == 4
    assert len(graph.edges) == 2
    assert ('state1', 'state2') in graph.edges
    assert ('state1', 'state3') in graph.edges


def test_networkx_renderer_render_invalid_engine() -> None:
    """Test that rendering an invalid engine raises an exception.

    The exception occurs because the engine is not properly initialized.
    """
    with pytest.raises(AttributeError):
        renderer = NetworkXRenderer(None)  # type: ignore
        renderer.render()



"""Sequence operations."""

from typing import Optional, Union

from nortl.core.protocols import Operand, Renderable
from nortl.core.renderers.operations.base import SequenceRenderer, const_unpack

__all__ = [
    'All',
    'Any',
    'Concat',
]


class Concat(SequenceRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        rendered_parts = [p.render(target) for p in self.parts]
        if len(rendered_parts) == 1:
            return rendered_parts[0]
        return f'{{{", ".join(rendered_parts)}}}'

    @staticmethod
    def eval(*args: Operand) -> Optional[Union[Renderable, int]]:
        """Evaluate operation into constant value or single operand.

        Unpacks the concatenation, if it consists of a single value.
        """
        args_ = tuple(const_unpack(arg) for arg in args)

        # Unpack single entry
        if len(args_) == 1:
            if (arg := args_[0]) is not None:
                return arg  # Constant value
            return args[0]  # Single renderable

        return None


class Any(SequenceRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        rendered_parts = [p.render(target) for p in self.parts]
        if len(rendered_parts) == 1:
            return rendered_parts[0]
        return f'({" || ".join(rendered_parts)})'

    @staticmethod
    def eval(*args: Operand) -> Optional[Union[Renderable, int]]:
        """Evaluate operation into constant value or single operand.

        Unpacks the Any, if it consists of a single value.
        Short circuits, if any element is > 0 or all are 0.
        """
        args_ = tuple(const_unpack(arg) for arg in args)

        # Unpack single entry
        if len(args_) == 1:
            if (arg := args_[0]) is not None:
                return arg  # Constant value
            return args[0]  # Single renderable

        # Short circuiting: any argument is > 0
        if any(arg is not None and arg > 0 for arg in args_):
            return 1

        # Short circuiting: all arguments are == 0
        if all(arg is not None and arg == 0 for arg in args_):
            return 0

        # Short circuiting: If "a" and "~a" are in the sequence
        filtered_args = [arg for arg in args if not isinstance(arg, (int, bool)) and arg is not None]

        rendered_args = [arg.render() for arg in filtered_args]
        for arg_ in filtered_args:
            if (~arg_).render() in rendered_args:
                return 1

        return None


class All(SequenceRenderer):
    def __call__(self, target: Optional[str] = None) -> str:
        rendered_parts = [p.render(target) for p in self.parts]
        if len(rendered_parts) == 1:
            return rendered_parts[0]
        return f'({" && ".join(rendered_parts)})'

    @staticmethod
    def eval(*args: Operand) -> Optional[Union[Renderable, int]]:
        """Evaluate operation into constant value or single operand.

        Unpacks the All, if it consists of a single value.
        Short circuits, if any element is 0 or all are > 0.
        """
        args_ = tuple(const_unpack(arg) for arg in args)

        # Unpack single entry
        if len(args_) == 1:
            if (arg := args_[0]) is not None:
                return arg  # Constant value
            return args[0]  # Single renderable

        # Short circuiting: any argument is == 0
        if any(arg is not None and arg == 0 for arg in args_):
            return 0

        # Short circuiting: all arguments are > 0
        if all(arg is not None and arg > 0 for arg in args_):
            return 1

        # Short circuiting: If "a" and "~a" are in the sequence
        filtered_args = [arg for arg in args if not isinstance(arg, (int, bool)) and arg is not None]

        rendered_args = [arg.render() for arg in filtered_args]
        for arg_ in filtered_args:
            if (~arg_).render() in rendered_args:
                return 0

        return None



from typing import Union

from nortl.core.protocols import EngineProto, ParameterProto, Renderable


class Timer:
    """This class represents a timer in the engine. It provides functions for operating the timer conveniently.

    It depends on the availability of a verilog module `nortl_count_down_timer`
    """

    def __init__(
        self, engine: EngineProto, width: Union[int, ParameterProto] = 16, instance_name_prefix: str = 'I_TIMER', clock_gating: bool = False
    ) -> None:
        """Initializes a timer object in the engine.

        This function instantiates a (count-down) timer module in the engine and holds the signals for the delay, reload and finish / zero.

        Arguments:
            engine (CoreEngine): Target engine object
            width (int): width of the counter register
            instance_name_prefix (str): Prefix for the instance name used in verilog. This also prefixes the names of the connected signals
            clock_gating (bool): If true, the unit is connected to the gated clock. If clock gating is disable, this setting is ignored.
        """
        self.engine = engine

        timer_idx = 0
        while (instance_name := f'{instance_name_prefix}_{timer_idx}') in engine.module_instances:
            timer_idx = timer_idx + 1

        self.instance_name = instance_name

        self.timer_module = self.engine.create_module_instance('nortl_count_down_timer', self.instance_name, clock_gating)

        self.engine.override_module_parameter(self.instance_name, 'DATA_WIDTH', width)

        self.delay = self.engine.define_local(f'{self.instance_name}_delay', reset_value=0, width=width)  # Todo: Could be an async signal
        self.reload = self.engine.define_local(f'{self.instance_name}_reload', reset_value=0)  # Todo: Could be an async signal
        self.zero = self.engine.define_local(f'{self.instance_name}_zero')  # Todo: Could be an async signal

        self.engine.connect_module_port(self.instance_name, 'DELAY', self.delay)
        self.engine.connect_module_port(self.instance_name, 'RELOAD', self.reload)
        self.engine.connect_module_port(self.instance_name, 'ZERO', self.zero)

    def wait_delay(self, delay: Union[Renderable, int, bool]) -> None:
        """Wait for a given delay without returning control to the engine. This can be seen as blocking delay.

        Arguments:
            delay (Renderable): Delay to wait for (in cycles)
        """
        self.start_delay(delay)
        self.engine.wait_for(self.finished)

    def start_delay(self, delay: Union[Renderable, int, bool]) -> None:
        """Start the timer with a given delay. This acts as a non-blocking delay and therefore returns control after starting the timer.

        Arguments:
            delay (Renderable): Delay to wait for (in cycles)
        """
        self.engine.set(self.delay, delay)
        self.engine.set(self.reload, 1)
        self.engine.sync()
        self.engine.set(self.reload, 0)
        self.engine.sync()

    @property
    def finished(self) -> Renderable:
        """Returns the signal (Renderable) that signals, if the timer has finished, i.e. the counter register is zero.

        Returns:
            Renderable: Zero-Flag of the counter
        """
        return self.zero



from .engine import CoreEngine
from .exceptions import (
    ConflictingAssignmentError,
    ExclusiveReadError,
    ExclusiveWriteError,
    ForbiddenAssignmentError,
    NonIdenticalRWError,
    OwnershipError,
    TransitionLockError,
    TransitionRestrictionError,
    UnfinishedForwardDeclarationError,
    WriteViolationError,
)
from .modifiers import Volatile
from .operations import All, Any, Concat, Const, IfThenElse, Var, to_renderable
from .parameter import Parameter
from .signal import Signal
from .tracing import enable_tracing

__all__ = [
    'All',
    'Any',
    'Concat',
    'ConflictingAssignmentError',
    'Const',
    'CoreEngine',
    'ExclusiveReadError',
    'ExclusiveWriteError',
    'ForbiddenAssignmentError',
    'IfThenElse',
    'NonIdenticalRWError',
    'OwnershipError',
    'Parameter',
    'Signal',
    'TransitionLockError',
    'TransitionRestrictionError',
    'UnfinishedForwardDeclarationError',
    'Var',
    'Volatile',
    'WriteViolationError',
    'enable_tracing',
    'to_renderable',
]



"""Tests for empty state removal optimization."""

from nortl import Engine


def test_empty_state_removal_basic() -> None:
    """Test basic empty state removal using collapse_sync context manager."""
    e = Engine('my_engine')

    # Engine has only the reset state at this point

    # Create three states that can be collapsed
    with e.collapse_sync():
        e.sync()
        e.sync()
        e.sync()

    # Introduce last state that will not be removed
    e.sync()
    e.current_state.name = 'Final State'

    # Run empty state removal
    e.empty_state_removal()

    # S_empty should be removed
    assert len(e.main_worker.states) == 2  # Now we only have two states

    cond, next_state = e.reset_state.transitions[0]

    assert len(e.reset_state.transitions) == 1
    assert cond.render() == "1'h1"
    assert next_state.name == 'Final State'


def test_empty_state_not_removed_with_assignments() -> None:
    """Test that states with assignments are not removed."""
    e = Engine('my_engine')

    with e.collapse_sync():
        e.sync()
        e.sync()
        e.set(e.define_scratch(1), 1)
        e.sync()

    e.sync()

    e.empty_state_removal()

    # State should not be removed because it has assignments
    assert len(e.main_worker.states) == 3



import pytest

from nortl import Engine
from nortl.core.constructs import Fork
from nortl.core.exceptions import ExclusiveReadError, ExclusiveWriteError, NonIdenticalRWError
from nortl.core.modifiers import Volatile


def test_same_read_same_write() -> None:
    """Tests, if the output values can actually be probed."""
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8)
    engine.set(output, 0)
    engine.sync()
    local_cnt = engine.define_local('local_counter', 8, reset_value=0)

    with Fork(engine, 'Timeouttimer') as proc:
        engine.set(local_cnt, 1)
        engine.sync()
        engine.set(local_cnt, local_cnt + 5)

        proc.finish()

    assert 'WORKER_1.Timeouttimer' in local_cnt.access_checker.reading_thread_names
    assert 'WORKER_1.Timeouttimer' in local_cnt.access_checker.writing_thread_names


def test_nonexcl_read() -> None:
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8)
    engine.set(output, 0)
    engine.sync()
    local_cnt = engine.define_local('local_counter', 8, reset_value=0)
    local_cnt2 = engine.define_local('local_counter2', 8, reset_value=0)

    with Fork(engine, 'Proc1') as proc1:
        engine.set(local_cnt2, local_cnt + 5)

        proc1.finish()

    with Fork(engine, 'Proc2') as proc2:
        with pytest.raises(ExclusiveReadError):
            engine.set(output, local_cnt + 2)

    del proc2


def test_nonexcl_write() -> None:
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8)
    engine.set(output, 0)
    engine.sync()
    local_cnt = engine.define_local('local_counter', 8, reset_value=0)

    with Fork(engine, 'Proc1') as proc1:
        engine.set(local_cnt, 1)
        engine.sync()
        engine.set(local_cnt, local_cnt + 5)

        proc1.finish()

    with Fork(engine, 'Proc2') as proc2:
        with pytest.raises(ExclusiveWriteError):
            engine.set(local_cnt, 2)

    del proc2


def test_nonid_read_write() -> None:
    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    output = engine.define_output('OUT', 8, reset_value=0)
    engine.sync()
    local_cnt = engine.define_local('local_counter', 8, reset_value=0)

    with Fork(engine, 'Proc1') as proc1:
        engine.set(local_cnt, 1)
        engine.sync()
        engine.set(local_cnt, 5)

        proc1.finish()

    with Fork(engine, 'Proc2') as proc2:
        with pytest.raises(NonIdenticalRWError):
            engine.set(output, local_cnt)

    del proc2


def test_indirection_via_combination_assignment() -> None:
    """Test that indirection through combinational assignments is caught.

    In the example, a signal is accessed from different threads, with one using a "indirection" through a combinational assignment.
    The access checker must not let this pass.
    """
    engine = Engine('my_engine')
    word = engine.define_input('word', width=16)
    lsb = engine.define_local('lsb', width=8, value=word[7:0])
    msb = engine.define_local('msb', width=8, value=word[15:8])

    lsb_out = engine.define_output('lsb_out', lsb.width, reset_value=0)
    msb_out = engine.define_output('msb_out', msb.width, reset_value=0)
    msb_out2 = engine.define_output('msb_out2', msb.width, reset_value=0)

    engine.sync()

    with Fork(engine, 'Proc1') as proc1:
        engine.set(lsb_out, lsb)
        engine.sync()
        proc1.finish()

    # Proc2 accesses word, after Proc1 already has
    with Fork(engine, 'Proc2'):
        with pytest.raises(ExclusiveReadError):
            engine.set(msb_out, msb)

    # Using a Volatile modifier allows the access
    with Fork(engine, 'Proc3') as proc3:
        engine.set(msb_out2, Volatile(msb))
        engine.sync()
        proc3.finish()


def test_free_after_join() -> None:
    """A signal accessed by a thread should be accessible without violaton after the thread has joined."""

    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    engine.sync()
    local_cnt = engine.define_local('local_counter', 8, reset_value=0)

    with Fork(engine, 'Proc1') as proc1:
        engine.set(local_cnt, 1)
        engine.sync()
        engine.set(local_cnt, 5)

        proc1.finish()

    engine.sync()

    proc1.join()

    engine.set(local_cnt, 1)


def test_handoff_to_thread() -> None:
    """A signal is written by the main thread before the fork and accessed inside of the fork."""

    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    engine.sync()
    local_cnt = engine.define_local('local_counter', 8, reset_value=0)
    engine.set(local_cnt, 1)

    with Fork(engine, 'Proc1') as proc1:
        engine.set(local_cnt, 1)
        engine.sync()
        engine.set(local_cnt, 5)

        proc1.finish()


def test_pass_to_thread() -> None:
    """A signal is written by the main thread before the fork and read inside of the fork."""

    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    engine.sync()
    local_cnt = engine.define_local('local_counter', 8, reset_value=0)
    engine.set(local_cnt, 1)
    dummy_before = engine.define_local('test', 2, 0)
    engine.set(dummy_before, local_cnt)

    with Fork(engine, 'proc2') as proc2:
        with Fork(engine, 'Proc1') as proc1:
            dummy = engine.define_scratch(2)
            engine.set(dummy, local_cnt)
            engine.sync()
            engine.set(local_cnt, 2)

        proc1.cancel()
        engine.set(local_cnt, 2)
        engine.set(dummy_before, local_cnt)

    proc2.cancel()
    dummy = engine.define_scratch(2)
    engine.set(dummy, local_cnt)


def test_pass_scratch_reg_to_thread() -> None:
    """A signal is written by the main thread before the fork and read inside of the fork."""

    engine = Engine('my_engine')
    _ = engine.define_input('IN')
    engine.sync()
    local_cnt = engine.define_scratch(4)
    engine.set(local_cnt, 1)
    dummy_before = engine.define_local('test', 2, 0)
    engine.set(dummy_before, local_cnt)

    with Fork(engine, 'proc2') as proc2:
        with Fork(engine, 'Proc1') as proc1:
            dummy = engine.define_scratch(2)
            engine.set(dummy, local_cnt)
            engine.sync()
            engine.set(local_cnt, 2)

        proc1.cancel()
        engine.set(local_cnt, 2)
        engine.set(dummy_before, local_cnt)

    proc2.cancel()
    dummy = engine.define_scratch(2)
    engine.set(dummy, local_cnt)



from nortl.core.checker import StaticAccessChecker
from nortl.core.common import NamedEntity, StaticAccess
from nortl.core.engine import CoreEngine
from nortl.core.manager.scratch_manager import ScratchManager
from nortl.core.manager.signal_manager import SignalManager
from nortl.core.module import Module, ModuleInstance
from nortl.core.operations import Const
from nortl.core.parameter import Parameter
from nortl.core.process import Thread, Worker
from nortl.core.protocols import (
    AssignmentProto,
    ConditionalAssignmentProto,
    EngineProto,
    ModuleInstanceProto,
    ModuleProto,
    NamedEntityProto,
    ParameterProto,
    ScratchManagerProto,
    ScratchSignalProto,
    SelectorAssignmentProto,
    SignalManagerProto,
    SignalProto,
    SignalSliceProto,
    StateProto,
    StaticAccessCheckerProto,
    StaticAccessProto,
    ThreadProto,
    WorkerProto,
)
from nortl.core.signal import ScratchSignal, Signal, SignalSlice
from nortl.core.state import Assignment, ConditionalAssignment, SelectorAssignment, State


def test_protocol_compatibility() -> None:
    """This test is used to debug the protocol<=>class mismatches. No real function, but good debugging."""
    t1: NamedEntityProto = NamedEntity('test')
    t2: ModuleInstanceProto = ModuleInstance(Module('test'), 'Test')
    t3: ModuleProto = Module('test')
    t4: EngineProto = CoreEngine('test')
    t5: StateProto = State(t4.main_worker, 'none')
    t6: SignalProto = Signal(t4, type='local', name='test', width=2)
    t7: SignalSliceProto = SignalSlice(t6, 1)
    t8: ScratchSignalProto = ScratchSignal(t6, 1)
    t9: ParameterProto = Parameter(t4, 'test', default_value=1)
    t10: WorkerProto = Worker(t4, 'test')
    t11: ThreadProto = Thread(t10, 'name')
    t12: StaticAccessProto = StaticAccess(t11)
    t13: StaticAccessCheckerProto = StaticAccessChecker(t6)
    t14: ScratchManagerProto = ScratchManager(t4)
    t15: SignalManagerProto = SignalManager(t4)
    t16: AssignmentProto = Assignment(t6, Const(1))
    t17: ConditionalAssignmentProto = ConditionalAssignment(t6, Const(1), Const(1))
    t18: SelectorAssignmentProto = SelectorAssignment(t6, {t8 == 1: Const(1)})

    del t1, t2, t3, t4, t5, t6, t7, t8, t9, t10, t11, t12, t13, t14, t15, t16, t17, t18



"""Modifiers can be applied to Renderables."""

from abc import abstractmethod
from contextlib import ExitStack
from typing import Generic, Literal, Never, Optional, Self, Set, TypeVar, Union

from nortl.core.exceptions import WriteViolationError
from nortl.core.operations import OperationTrait
from nortl.core.protocols import (
    ACCESS_CHECKS,
    SIGNAL_ACCESS_CHECKS,
    AnySignal,
    AssignmentTarget,
    EngineProto,
    PermanentSignal,
    Renderable,
)

__all__ = [
    'UnregisteredRead',
    'Volatile',
    'WeakReference',
]

T_Content = TypeVar('T_Content', bound=Renderable)
T_Signal = TypeVar('T_Signal', bound=AnySignal)
T_PermanentSignal = TypeVar('T_PermanentSignal', bound=PermanentSignal)


# Modifiers
class BaseModifier(Generic[T_Content], OperationTrait, ExitStack):
    """Baseclass for Modifiers."""

    def __init__(self, content: T_Content) -> None:
        """Initialize an alias.

        Arguments:
            content: An operation result.
        """
        self._content = content
        super().__init__()

    @property
    def content(self) -> T_Content:
        """Content of the alias."""
        return self._content

    @abstractmethod
    def copy(self, content: T_Content) -> Self:
        """Copy modifier to new content."""

    # Implement OperationTrait
    @property
    def is_primitive(self) -> bool:
        """Indicates if this object is a Verilog primitive."""
        return self.content.is_primitive

    @property
    def is_constant(self) -> bool:
        """Indicates if this object has a constant value."""
        return self.content.is_constant

    @property
    def operand_width(self) -> Optional[int]:
        """Indicates the width when used as an operand, equal to the width of the alias operation.

        A width of None means that the width is not fixed during execution of noRTL.
        """
        return self.content.operand_width

    def render(self, target: Optional[str] = None) -> str:
        """Render operation to target language.

        Arguments:
            target: Target language.
        """
        return self.content.render(target=target)

    def __format__(self, format_spec: str) -> str:
        return self.render()

    # TODO Modifier must only be valid within current construct

    # Support unpacking of constants
    @property
    def value(self) -> int:
        """Value of content, if the object has a constant value."""
        if self.is_constant:
            return self.content.value  # type: ignore[attr-defined, no-any-return]
        else:
            raise RuntimeError("Content of modifier is not constant. It's value must not be accessed.")


class UnregisteredRead(Generic[T_Content], BaseModifier[T_Content]):
    """Modifier for signals or operations, that hides them from read accesses.

    Any read access to signals inside a UnregisteredRead modifier is not registered.
    Note that this does not permanently disable the access checks, when acessing the signals normally.

    !!! danger
        This modifier is meant for internal purposes. For example, it is used in the noRTL test wrapper for assertions.
        It must not be used in production code!
    """

    def copy(self, content: T_Content) -> 'UnregisteredRead[T_Content]':
        """Copy modifier to new content."""
        return UnregisteredRead(content)

    def read_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register read access from the current thread, state and construct depth.

        Does nothing for UnregisteredRead.
        """


class Volatile(Generic[T_Signal], BaseModifier[T_Signal]):
    """Modifier for signals, allowing them to be shared between threads.

    Wrapping a noRTL signal will temporarily ignore the following checks by default:

    - Exclusive Read Check ([ExclusiveReadError][nortl.core.exceptions.ExclusiveReadError]): The modified object can be read from multiple threads.
    - Non-Identical Read/Write Check ([NonIdenticalRWError][nortl.core.exceptions.NonIdenticalRWError]): The modified object can be read/written from different threads.

    It's possible to select other checks to ignore, by listing their names, e.g.:

    ```python
    # Ignore a set of checks
    unsafe_signal = Volatile(signal, 'identical_rw', 'exclusive_write')

    # Ignore a single check
    unsafe_signal = Volatile(signal, 'identical_rw')
    ```

    Note that this does not permanently disable the access checks. The read or write access to the signal is still registered.
    If the same signal is used in other places, it may still cause access violations exceptions.
    """

    def __init__(self, content: T_Signal, *ignore: SIGNAL_ACCESS_CHECKS) -> None:
        super().__init__(content)
        if len(ignore) > 0:
            self._ignore: Set[SIGNAL_ACCESS_CHECKS] = set(ignore)
        else:
            self._ignore = {'exclusive_read', 'identical_rw'}

    @property
    def ignore(self) -> Set[SIGNAL_ACCESS_CHECKS]:
        """Set of ignored access checks."""
        return self._ignore

    def copy(self, content: T_Signal) -> 'Volatile[T_Signal]':
        """Copy modifier to new content."""
        return Volatile(content, *self.ignore)

    # Implement OperationTrait
    def read_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register read access from the current thread, state and construct depth."""
        self.content.read_access(ignore=ignore | self.ignore)

    # Implement AssignmentTarget
    @property
    def name(self) -> str:
        """Just return name."""
        return self.content.name

    @property
    def engine(self) -> EngineProto:
        """NoRTL engine that the signal belongs to."""
        return self.content.engine

    def write_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register write access from the current thread."""
        self.content.write_access(ignore=ignore | self.ignore)

    def overlaps_with(self, other: AssignmentTarget) -> Union[bool, Literal['partial']]:
        """Check if signal overlaps with other signal or signal slice."""
        return self.content.overlaps_with(other)


class ReadOnly(Generic[T_Signal], BaseModifier[T_Signal]):
    """Modifier for signals to mark them as read-only.

    Writing to a noRTL signal that is wrapped in a ReadOnly modifier will throw an Access Violation.
    """

    def copy(self, content: T_Signal) -> 'ReadOnly[T_Signal]':
        """Copy modifier to new content."""
        return ReadOnly(content)

    # Implement OperationTrait
    def read_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register read access from the current thread, state and construct depth."""
        self.content.read_access(ignore=ignore)

    # Implement AssignmentTarget
    @property
    def name(self) -> str:
        """Just return name."""
        return self.content.name

    @property
    def engine(self) -> EngineProto:
        """NoRTL engine that the signal belongs to."""
        return self.content.engine

    def write_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> Never:
        """Register write access from the current thread."""
        raise WriteViolationError(f'Signal {self.name} is marked as read-only.')

    def overlaps_with(self, other: AssignmentTarget) -> Union[bool, Literal['partial']]:
        """Check if signal overlaps with other signal or signal slice."""
        return self.content.overlaps_with(other)

    # Emulate Scratch Signal
    def __enter__(self) -> Self:
        if hasattr(self.content, '__enter__'):
            self.enter_context(self.content)  # type: ignore
        return self


class WeakReference(Generic[T_PermanentSignal], BaseModifier[T_PermanentSignal]):
    """Modifier for signals to mark them as weak references to the underlying signal.

    The reference can be expired at any time.
    """

    def __init__(self, content: T_PermanentSignal):
        super().__init__(content)
        self._expired = False

    def copy(self, content: T_PermanentSignal) -> 'WeakReference[T_PermanentSignal]':
        """Copy modifier to new content."""
        return WeakReference(self.content)

    @property
    def expired(self) -> bool:
        """Indicates if this signal is expired."""
        return self._expired

    def expire(self) -> None:
        """Marks this signal as expired. Any further access is forbidden."""
        self._expired = True

    # Implement OperationTrait
    def read_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register read access from the current thread, state and construct depth."""
        if self.expired:
            raise RuntimeError(
                f'Signal {self.name} is expired.\nThis can happen, if the signal was returned from a segment and the segment was called again, making this reference to the signal invalid.'
            )
        self.content.read_access(ignore=ignore)

    # Implement AssignmentTarget
    @property
    def name(self) -> str:
        """Just return name."""
        return self.content.name

    @property
    def engine(self) -> EngineProto:
        """NoRTL engine that the signal belongs to."""
        return self.content.engine

    def write_access(self, ignore: Set[ACCESS_CHECKS] = set()) -> None:
        """Register write access from the current thread."""
        if self.expired:
            raise RuntimeError(f'Signal {self.name} is expired.')
        self.content.write_access(ignore=ignore)

    def overlaps_with(self, other: AssignmentTarget) -> Union[bool, Literal['partial']]:
        """Check if signal overlaps with other signal or signal slice."""
        return self.content.overlaps_with(other)



# ruff: noqa: N801

from nortl.components.channel import Channel, ElasticChannel
from nortl.core import CoreEngine
from nortl.utils.test_wrapper import NoRTLTestBase


class TestChannel(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> CoreEngine:
        engine = CoreEngine('my_engine')
        self.channel = Channel(engine, 8, 'testchannel')
        return engine

    def dut(self, engine: CoreEngine) -> None:
        for i in range(10):
            self.channel.send(i)

    def the_testbench(self, engine: CoreEngine) -> None:
        for i in range(10):
            with self.channel.receive() as rec_val:
                self.assertTrue(rec_val == i)


class TestChannelBlocksSender(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> CoreEngine:
        self.expect_timeout = 1
        engine = CoreEngine('my_engine')
        self.channel = Channel(engine, 8, 'testchannel')
        return engine

    def dut(self, engine: CoreEngine) -> None:
        for i in range(10):
            self.channel.send(i)
        self.assertTrue(0)

    def verify_final_state(self, engine: CoreEngine) -> None:
        self.assertTrue(engine.signals['timeout'])
        self.finish_simulation()


class TestEmptyChannelBlocksReciever(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> CoreEngine:
        self.expect_timeout = 1
        engine = CoreEngine('my_engine')
        self.channel = Channel(engine, 8, 'testchannel')
        return engine

    def dut(self, engine: CoreEngine) -> None:
        for i in range(10):
            self.channel.receive()
        self.assertTrue(0)

    def verify_final_state(self, engine: CoreEngine) -> None:
        self.assertTrue(engine.signals['timeout'])
        self.finish_simulation()


class TestElasticChannel_Trivial(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> CoreEngine:
        engine = CoreEngine('my_engine')
        self.channel = ElasticChannel(engine, 8, 'testchannel', 8)
        return engine

    def dut(self, engine: CoreEngine) -> None:
        for i in range(10):
            self.channel.send(i)

    def the_testbench(self, engine: CoreEngine) -> None:
        for i in range(10):
            with self.channel.receive() as rec_val:
                self.assertTrue(rec_val == i)


class TestElasticChannel_SendMultiple(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> CoreEngine:
        engine = CoreEngine('my_engine')
        self.channel = ElasticChannel(engine, 8, 'testchannel', 8)
        return engine

    def dut(self, fsm: CoreEngine) -> None:
        for i in range(5):
            print(2 * i, 2 * i + 1)
            self.channel.send_multiple([2 * i + 1, 2 * i])

    def the_testbench(self, fsm: CoreEngine) -> None:
        for i in range(10):
            with self.channel.receive() as rec_val:
                self.assertEqual(rec_val, i)


class TestElasticChannel_sender_faster_than_rec(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> CoreEngine:
        engine = CoreEngine('my_engine')
        self.channel = ElasticChannel(engine, 8, 'testchannel', 8)
        return engine

    def dut(self, engine: CoreEngine) -> None:
        for i in range(10):
            self.channel.send(i)

    def the_testbench(self, engine: CoreEngine) -> None:
        for i in range(10):
            engine.sync()
            with self.channel.receive() as rec_val:
                self.assertTrue(rec_val == i)


class TestElasticChannel_delayed_rec(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> CoreEngine:
        engine = CoreEngine('my_engine')
        self.channel = ElasticChannel(engine, 8, 'testchannel', 8)
        return engine

    def dut(self, engine: CoreEngine) -> None:
        for i in range(100):
            self.channel.send(i)

    def the_testbench(self, engine: CoreEngine) -> None:
        for _ in range(256):
            engine.sync()
        for i in range(100):
            with self.channel.receive() as rec_val:
                self.assertTrue(rec_val == i)


class TestElasticChannel_delayed_send(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> CoreEngine:
        engine = CoreEngine('my_engine')
        self.channel = ElasticChannel(engine, 8, 'testchannel', 8)
        return engine

    def dut(self, engine: CoreEngine) -> None:
        for _ in range(256):
            engine.sync()
        for i in range(100):
            self.channel.send(i)

    def the_testbench(self, engine: CoreEngine) -> None:
        for i in range(100):
            with self.channel.receive() as rec_val:
                self.assertTrue(rec_val == i)


class TestElasticChannel_rec_faster_than_sender(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> CoreEngine:
        engine = CoreEngine('my_engine')
        self.channel = ElasticChannel(engine, 8, 'testchannel', 8)
        return engine

    def dut(self, engine: CoreEngine) -> None:
        engine.sync()
        for i in range(10):
            self.channel.send(i)
        engine.sync()

    def the_testbench(self, engine: CoreEngine) -> None:
        for i in range(10):
            with self.channel.receive() as rec_val:
                self.assertTrue(rec_val == i)


class TestElasticChannelBlocksSender(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> CoreEngine:
        self.expect_timeout = 1
        engine = CoreEngine('my_engine')
        self.channel = ElasticChannel(engine, 8, 'testchannel', 8)
        return engine

    def dut(self, engine: CoreEngine) -> None:
        for i in range(10):
            self.channel.send(i)
        self.assertTrue(0)

    def verify_final_state(self, engine: CoreEngine) -> None:
        self.assertTrue(engine.signals['timeout'])
        self.finish_simulation()


class TestEmptyElasticChannelBlocksReciever(NoRTLTestBase[CoreEngine]):
    def init_sequence(self) -> CoreEngine:
        self.expect_timeout = 1

        engine = CoreEngine('my_engine')
        self.channel = ElasticChannel(engine, 8, 'testchannel', 8)
        return engine

    def dut(self, engine: CoreEngine) -> None:
        for i in range(10):
            self.channel.receive()
        self.assertTrue(0)

    def verify_final_state(self, engine: CoreEngine) -> None:
        self.assertTrue(engine.signals['timeout'])
        self.finish_simulation()



from nortl.renderer.verilog_utils.structural import VerilogDeclaration


def test_register_declaration_width_0() -> None:
    target = VerilogDeclaration('logic', 'A')
    assert target.render() == 'logic A'

    target = VerilogDeclaration('logic', 'A', 0)
    assert target.render() == 'logic A'


def test_register_declaration() -> None:
    target = VerilogDeclaration('logic', 'A', 4)
    assert target.render() == 'logic [3:0] A'


def test_double_register_declaration() -> None:
    target = VerilogDeclaration('logic', ['A', 'B'], 4)
    assert target.render() == 'logic [3:0] A, B'


def test_enum() -> None:
    target = VerilogDeclaration('enum logic', 'A', 4, members=['STATE1', 'STATE2'])
    assert target.render() == 'enum logic [3:0] {STATE1, STATE2} A'

    target = VerilogDeclaration('enum logic', 'A')
    target.add_member('STATE1', 0)
    target.add_member('STATE2', 1)

    assert target.render() == 'enum logic {STATE1 = 0, STATE2 = 1} A'


def test_param() -> None:
    target = VerilogDeclaration('BU', 'A')
    target.add_parameter('WIDTH', 1)

    assert target.render() == 'BU #(.WIDTH(1)) A ()'


def test_connection() -> None:
    target = VerilogDeclaration('BU', 'A')
    target.add_connection('A', 'B')
    target.add_connection('C', 'D')

    assert target.render() == 'BU A (.A(B), .C(D))'
